/*
 * stalk/interpret.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "interpret.h"
#include "var.h"
#include "stuff.h"
#include "functions.h"
#include "task.h"
#include "debug.h"

int debug_dump;

static char *mnems[] = {
  "none", "noop", "jump",
  "if", "elseif", "else", "endif",
  "while", "endwhile",
  "exit", "return", "return (empty)",
  "for (list)", "for (range)", "endfor",
  "fork", "endfork",
  "identifier", "variable (local)", "variable (global)",
  "value", "end expression",
  "element", "element (range)",
  "single element list", "combine lists", "splice lists", "empty list", "list",
  "assign (local)", "assign (global)", "end assign",
  "function", "function (indirect)", "<x> ? <y> : <z>",
  "equal", "not equal",
  "greater than or equal", "greater than", "less than or equal", "less than",
  "add", "subtract", "multiply", "divide", "modulus", "negate",
  "or", "and", "not", "or (short circuit)", "and (short circuit)", "in",
  "line info",
  "debug", "end debug", "call", "ignore", "fatal"
};

typedef void (*operation) P((task *t));

static operation operation_table[] = {
  ntrp_noop, ntrp_noop, ntrp_jump,
  ntrp_if, ntrp_if, ntrp_noop, ntrp_noop,
  ntrp_while, ntrp_endwhile,
  ntrp_exit, ntrp_return, ntrp_empty_return,
  ntrp_for_list, ntrp_for_range, ntrp_endfor,
  ntrp_fork, ntrp_endfork,
  ntrp_identifier, ntrp_local_variable, ntrp_global_variable,
  ntrp_value, ntrp_end_expression,
  ntrp_element, ntrp_element_range,
  ntrp_single_elt_list, ntrp_combine_lists, ntrp_splice_lists,
  ntrp_empty_list, ntrp_noop,
  ntrp_local_assign, ntrp_global_assign, ntrp_end_assign,
  ntrp_function, ntrp_function, ntrp_conditional,
  ntrp_equal, ntrp_not_equal,
  ntrp_greater_equal, ntrp_greater_than, ntrp_less_equal, ntrp_less_than,
  ntrp_add, ntrp_sub, ntrp_mul, ntrp_div, ntrp_mod, ntrp_negate,
  ntrp_or, ntrp_and, ntrp_not, ntrp_or_ss, ntrp_and_ss, ntrp_in,
  ntrp_line_info,
  ntrp_debug, ntrp_end_debug, ntrp_debug_call, ntrp_debug_ignore,
  ntrp_debug_fatal,
  NULL
};

var *in_pop(t)
     task *t;
{
  var_stack *stack;
  var *v;

  if (!t->stack)
    die("%s: %s, %s %d\n  %s: %s\n  %s %d: %s\n  %s: %x\n  %s: %s\n",
	"task.c", "interpreter stack underflow",
	"line", __LINE__,
	"program", t->program->name,
	"program line", t->line_info->list[0]->num,
	t->line_info->list[1]->string,
	"task address", t,
	"mnemonic", mnems[t->current->type]);
  stack = t->stack;
  t->stack = t->stack->next;
  v = stack->val;
  free(stack);
  return v;
}

void in_push(t, v)
     task *t;
     var *v;
{
  var_stack *stack;
  stack = NEW(var_stack);
  stack->val = v;
  stack->next = t->stack;
  t->stack = stack;
}

void ntrp_task(t)
     task *t;
{
#ifdef EVIL_DEBUGGING_MODE
  if (debug_dump && !t->dumpfile) {
    static int num = 0;
    char name[20];

    dump_program(t->current);
    sprintf(name, "stackfile%d", ++num);
    if ((t->dumpfile = fopen(name, "w")) == NULL)
      printf("couldn't open stackfile. sorry 'bout your luck. errno == %d\n",
	     errno);
  }
#endif /* EVIL_DEBUGGING_MODE */

  while ((t->type == TA_READY) && t->current) {
    if (t->error_pending) {
      debug_task(t, t->error);
      if (t->type != TA_READY)
	return;
    }

    if (t->counter > 0)
      if (--t->counter == 0) {
	debug_task(t, E_TICKS);
	if (t->type != TA_READY)
	  return;
      }

#ifdef EVIL_DEBUGGING_MODE
    if (debug_dump && t->dumpfile) {
      fprintf(t->dumpfile, "about to exec %s\n", mnems[t->current->type]);
      fprintf(t->dumpfile, "current stack:\n");
      dump_stack(t->dumpfile, t->stack);
      fprintf(t->dumpfile, "* end dump *\n\n");
    }
#endif /* EVIL_DEBUGGING_MODE */

    /*
     * look the function pointer up in the operation table
     * and execute it
     */
    (operation_table[(int) t->current->type])(t);
  }

  if (!t->current) {
    in_push(t, NULL);
    t->type = TA_RETURNING;
  }
}

#ifdef EVIL_DEBUGGING_MODE
void dump_program(s)
     statement *s;
{
  char name[20], *c;
  static int num = 0;
  FILE *f;

  num++;
  sprintf(name, "programfile%d", num);
  if ((f = fopen(name, "w")) == NULL) {
    printf("couldn't open programfile. sorry 'bout your luck. errno == %d\n",
	   errno);
    return;
  }
  while (s) {
    fprintf(f, "address %x\n", (int) s);
    fprintf(f, "  mnemonic %s\n", mnems[s->type]);
    fprintf(f, "  next %x\n", (int) s->next);
    if (s->jump)
      fprintf(f, "  jump %x\n", (int) s->jump);
    if (s->type == MN_ID || s->type == MN_VALUE) {
      c = vartostr(s->value);
      fprintf(f, "  value %s\n", c);
      free(c);
    } else if (s->type == MN_LINE) {
      fprintf(f, "  line %d: %s\n", s->value->list[0]->num,
	      s->value->list[1]->string);
    }
    s = s->next;
  }
  fclose(f);
}

void dump_stack(f, vs)
     FILE *f;
     var_stack *vs;
{
  if (!vs)
    return;

  dump_stack(f, vs->next);
    dump_var(f, vs->val, 0);
}

void dump_var(f, v, indent)
     FILE *f;
     var *v;
     int indent;
{
  int x;

  for (x=0; x<indent; x++)
    fprintf(f, "  ");
  if (v == NULL)
    fprintf(f, "NULL\n");
  else
    switch (v->type) {
    case V_STRING:
      fprintf(f, "STRING: %s\n", v->string);
      break;
    case V_LABEL:
      fprintf(f, "LABEL: %s\n", v->string);
      break;
    case V_TOKEN:
      fprintf(f, "TOKEN: %s\n", mnems[v->num]);
      break;
    case V_ERR:
      fprintf(f, "ERROR: %s\n", err2string(v->err));
      break;
    case V_NUM:
      fprintf(f, "NUM: %d\n", v->num);
      break;
    case V_INDEX:
      fprintf(f, "INDEX: %d\n", v->num);
      break;
    case V_LIST:
      if (v->num == 1)
	fprintf(f, "LIST. 1 element:\n");
      else
	fprintf(f, "LIST. %d elements:\n", v->num);
      for (x=0; x<v->num; x++)
	dump_var(f, v->list[x], indent+1);
      break;
    default:
      fprintf(f, "invalid variable type %d\n", v->type);
      break;
    }
  fflush(f);
}
#endif /* EVIL_DEBUGGING_MODE */

/*
 * Duh. a no-op. a null operation. do nothing.
 * no stack activity here...
 */
void ntrp_noop(t)
     task *t;
{
  t->current = t->current->next;
}

/*
 * jump to the statement pointed to by t->current->jump
 * used in places like the end of a while (to jump back to the statement(s)
 * that construct the condition value, to jump to the endif at the end of
 * a successful if, etc.
 * no stack activity here...
 */
void ntrp_jump(t)
     task *t;
{
  t->current = t->current->jump;
}

/*
 * when this mnemonic is called, the result of the condition should be on
 * top of the stack. pops the result off the stack. If true, executing
 * continues with the next instruction. If false, the execution jumps to
 * the statement pointed to by the jump field of the current statement.
 * this handles if and elseif statements.
 */
void ntrp_if(t)
     task *t;
{
  var *cond;

  cond = in_pop(t);
  if (!istrue(cond))
    t->current = t->current->jump;
  else
    t->current = t->current->next;
  var_free(cond);
}

/*
 * the result of the condition should be on the stack. pop the condition off.
 * if true, continue with the body. if false, jump to after the endwhile.
 * this will push a sentinel label onto the stack so that an exit statement
 * may unwind the stack properly.
 */
void ntrp_while(t)
     task *t;
{
  var *cond;

  cond = in_pop(t);
  if (!istrue(cond))
    t->current = t->current->jump->next;
  else {
    /* push that sentinel onto the stack for exit statements */
    in_push(t, var_ref(t->current->value));
    t->current = t->current->next;
  }
  var_free(cond);
}

/*
 * pop the 'exit' label off the stack that was pushed on by the corresponding
 * while mnemonic. jump back to the condition for the while
 */
void ntrp_endwhile(t)
     task *t;
{
  var_free(in_pop(t));
  t->current = t->current->jump;
}

/*
 * unwind the stack until we get to the appropriate sentinel label
 * put on the stack by the while/for statement that we're exiting from,
 * then jump to after the endwhile/endfor
 */
void ntrp_exit(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  while (v->type != V_LABEL || strcmp(v->string, t->current->value->string)) {
    var_free(v);
    v = in_pop(t);
  }
  var_free(v);

  /*
   * jump past the endwhile/endfor and the associated line mnemonic
   */
  t->current = t->current->jump->next->next;
}

/*
 * the return value will be on top of the stack.
 * mark the task as returning so that the scheduler can give the return
 * value to the parent, destroy this task's storage, and wake the parent.
 */
void ntrp_return(t)
     task *t;
{
  t->type = TA_RETURNING;
}

/*
 * same as return, but we have to put a NULL value on top of the stack
 */
void ntrp_empty_return(t)
     task *t;
{
  in_push(t, NULL);
  t->type = TA_RETURNING;
}

/*
 * the top of the stack will look like this:
 *
 * --> list    (the list we are iterating over)
 *     index   (the index of the next item we are to extract from the list
 *              using C's start from zero convention)
 *     name    (the name of the local variable that gets the value)
 *     ...
 *
 * either get the next element, put it into the var, and restore the stack
 * before continuing with the body of the for, or pop these 3 items off the
 * stack and jump to the statement after the endfor
 *
 * when initially set up, the stack will have a num as index. to keep
 * the for loop from destroying the value of the variable (in case it is
 * a reference to some variable in use elsewhere), the first time
 * the for loop executes, the index is copied and saved back onto the
 * stack as a token
 */
void ntrp_for_list(t)
     task *t;
{
  var *idx, *list, *name, *temp;

  list = in_pop(t);
  idx = in_pop(t);
  name = in_pop(t);

  /* copy index if this is the first iteration */
  if (idx->type == V_NUM) {
    temp = newnum(idx->num);
    temp->type = V_INDEX;
    var_free(idx);
    idx = temp;
  }

  if (idx->type != V_INDEX || list->type != V_LIST || name->type != V_LABEL) {
    raise_error(t, E_TYPE);
    var_free(name);
    var_free(idx);
    var_free(list);
    t->current = t->current->jump;
  } else if (idx->num >= list->num) {
    var_free(name);
    var_free(idx);
    var_free(list);
    t->current = t->current->jump;
  } else {
    t->local = set_var(t->local, name->string, list->list[idx->num]);
    idx->num++;
    in_push(t, name);
    in_push(t, idx);
    in_push(t, list);
  }

  /* we want to do this no matter which of the above cases happened */
  t->current = t->current->next;
}

/*
 * the top of the stack will look like this:
 *
 * --> endval  (the ending value for the range)
 *     newval  (the value that 'name' is supposed to receive)
 *     name    (the name of the local variable that gets the value)
 *     ...
 *
 * either get the next element, put it into the var, and restore the stack
 * before continuing with the body of the for, or pop these 3 items off the
 * stack and jump to the statement after the endfor
 *
 * when initially set up, the stack will have a num as newval. to keep
 * the for loop from destroying the value of the variable (in case it is
 * a reference to some variable in use elsewhere), the first time
 * the for loop executes, the newval is copied and saved back onto the
 * stack as a token
 */
void ntrp_for_range(t)
     task *t;
{
  var *endval, *newval, *name, *temp;

  endval = in_pop(t);
  newval = in_pop(t);
  name = in_pop(t);

  /* copy newval if this is the first iteration */
  if (newval->type == V_NUM) {
    temp = newnum(newval->num);
    temp->type = V_INDEX;
    var_free(newval);
    newval = temp;
  }

  if (newval->type != V_INDEX || endval->type != V_NUM ||
      name->type != V_LABEL) {
    raise_error(t, E_TYPE);
    var_free(endval);
    var_free(newval);
    var_free(name);
  } else if (newval->num > endval->num) {
    var_free(endval);
    var_free(newval);
    var_free(name);
    t->current = t->current->jump;
  } else {
    t->local = set_var(t->local, name->string, newval);
    newval->num++;
    in_push(t, name);
    in_push(t, newval);
    in_push(t, endval);
  }

  /*
   * we want to do this no matter which of the above cases happened.
   * in the case of the above jump, the jump takes us to the endfor, which
   * we now want to skip
   */
  t->current = t->current->next;
}

/*
 * jump back to the 'for' mnemonic for the next iteration
 */
void ntrp_endfor(t)
     task *t;
{
  t->current = t->current->jump;
}

/*
 * start a new task that starts with the next statement.
 * the top of the stack holds the number of seconds to wait before it starts.
 * continue this task after the endfork.
 */
void ntrp_fork(t)
     task *t;
{
  var *seconds;

  seconds = in_pop(t);
  if (seconds->type == V_NUM)
    fork_task(t->buf, t->program, t->current->next, seconds->num);
  else
    raise_error(t, E_TYPE);
  var_free(seconds);
  t->current = t->current->jump->next;
}

/*
 * if a task gets to this statement, it has been forked and has reached the end
 * of the section it's supposed to execute
 */
void ntrp_endfork(t)
     task *t;
{
  t->type = TA_DYING;
}

/*
 * push the identifier onto the stack. this mnemonic should be followed by
 * a local_var mnemonic or something which tasks the identifier and uses
 * it
 */
void ntrp_identifier(t)
     task *t;
{
  in_push(t, var_ref(t->current->value));
  t->current = t->current->next;
}

/*
 * take the identifier that should have been pushed onto the top of the stack
 * and use it as the name of the local variable to look up. push the local
 * var's value onto the stack
 */
void ntrp_local_variable(t)
     task *t;
{
  var *name, *val;

  name = in_pop(t);
  in_push(t, val = get_var(t->local, name->string));
  if (val->type == V_ERR)
    raise_error(t, val->err);
  var_free(name);
  t->current = t->current->next;
}

/*
 * take the identifier that should have been pushed onto the top of the stack
 * and use it as the name of the global variable to look up. push the global
 * var's value onto the stack
 */
void ntrp_global_variable(t)
     task *t;
{
  var *name, *val;

  name = in_pop(t);
  in_push(t, val = get_var(t->buf->global, name->string));
  if (val->type == V_ERR)
    raise_error(t, val->err);
  var_free(name);
  t->current = t->current->next;
}

/*
 * push the value onto the stack
 */
void ntrp_value(t)
     task *t;
{
  in_push(t, var_ref(t->current->value));
  if (t->current->value && (t->current->value->type == V_ERR))
    raise_error(t, t->current->value->err);
  t->current = t->current->next;
}

/*
 * the result of the expression should be on the stack. pop and discard it.
 */
void ntrp_end_expression(t)
     task *t;
{
  var_free(in_pop(t));
  t->current = t->current->next;
}

void ntrp_element(t)
     task *t;
{
  var *name, *idx;

  idx = in_pop(t); /* the index */
  name = in_pop(t); /* the variable name/token (if assignment) */
  if (name->type == V_TOKEN) {
    /* we're doing an assignment statement */
    var *lhs, *i1, *i2, *val;

    if (name->num == MN_LOCAL_ASSIGN) {
      lhs = in_pop(t);
      val = get_var(t->local, lhs->string);
      if (val->type == V_ERR)
	raise_error(t, val->err);
      in_push(t, lhs);
      in_push(t, name);
      if ((val->type != V_STRING) && (val->type != V_LIST)) {
	in_push(t, make_error(t, E_TYPE));
	var_free(val);
      } else {
	in_push(t, val);
      }
    } else if (name->num == MN_GLOBAL_ASSIGN) {
      lhs = in_pop(t);
      val = get_var(t->buf->global, lhs->string);
      if (val->type == V_ERR)
	raise_error(t, val->err);
      in_push(t, lhs);
      in_push(t, name);
      if ((val->type != V_STRING) && (val->type != V_LIST)) {
	in_push(t, make_error(t, E_TYPE));
	var_free(val);
      } else {
	in_push(t, val);
      }
    } else if (name->num == MN_ELEMENT) {
      i1 = in_pop(t);
      lhs = in_pop(t);
      if (i1->type != V_NUM)
	val = make_error(t, E_TYPE);
      else
	val = extract_elements(t, lhs, i1->num, -1);
      in_push(t, lhs);
      in_push(t, i1);
      in_push(t, name);
      in_push(t, val);
    } else if (name->num == MN_ELEMENT_RANGE) {
      i2 = in_pop(t);
      i1 = in_pop(t);
      lhs = in_pop(t);
      if ((i1->type != V_NUM) || (i2->type != V_NUM))
	val = make_error(t, E_TYPE);
      else
	val = extract_elements(t, lhs, i1->num, i2->num);
      in_push(t, lhs);
      in_push(t, i1);
      in_push(t, i2);
      in_push(t, name);
      in_push(t, val);
    }
    in_push(t, idx);
    in_push(t, newtoken(MN_ELEMENT));
  } else {
    if (idx->type != V_NUM)
      in_push(t, make_error(t, E_TYPE));
    else
      in_push(t, extract_elements(t, name, idx->num, -1));
    var_free(name);
    var_free(idx);
  }
  t->current = t->current->next;
}

void ntrp_element_range(t)
     task *t;
{
  var *lidx, *ridx, *name;

  ridx = in_pop(t); /* the right index */
  lidx = in_pop(t); /* the left index */
  name = in_pop(t); /* the variable name/token (if assignment) */
  if (name->type == V_TOKEN) {
    /* we're doing an assignment statement */
    var *lhs, *i1, *i2, *val;

    if (name->num == MN_LOCAL_ASSIGN) {
      lhs = in_pop(t);
      val = get_var(t->local, lhs->string);
      if (val->type == V_ERR)
	raise_error(t, val->err);
      in_push(t, lhs);
      in_push(t, name);
      if ((val->type != V_STRING) && (val->type != V_LIST)) {
	in_push(t, make_error(t, E_TYPE));
	var_free(val);
      } else {
	in_push(t, val);
      }
    } else if (name->num == MN_GLOBAL_ASSIGN) {
      lhs = in_pop(t);
      val = get_var(t->buf->global, lhs->string);
      if (val->type == V_ERR)
	raise_error(t, val->err);
      in_push(t, lhs);
      in_push(t, name);
      if ((val->type != V_STRING) && (val->type != V_LIST)) {
	in_push(t, make_error(t, E_TYPE));
	var_free(val);
      } else {
	in_push(t, val);
      }
    } else if (name->num == MN_ELEMENT) {
      i1 = in_pop(t);
      lhs = in_pop(t);
      if (i1->type != V_NUM)
	val = make_error(t, E_TYPE);
      else
	val = extract_elements(t, lhs, i1->num, -1);
      in_push(t, lhs);
      in_push(t, i1);
      in_push(t, name);
      in_push(t, val);
    } else if (name->num == MN_ELEMENT_RANGE) {
      i2 = in_pop(t);
      i1 = in_pop(t);
      lhs = in_pop(t);
      if ((i1->type != V_NUM) || (i2->type != V_NUM))
	val = make_error(t, E_TYPE);
      else
	val = extract_elements(t, lhs, i1->num, i2->num);
      in_push(t, lhs);
      in_push(t, i1);
      in_push(t, i2);
      in_push(t, name);
      in_push(t, val);
    }
    in_push(t, lidx);
    in_push(t, ridx);
    in_push(t, newtoken(MN_ELEMENT_RANGE));
  } else {
    if (lidx && ridx && (((lidx->type != V_NONE) && (lidx->type != V_NUM))
			 || ((ridx->type != V_NONE) && (ridx->type != V_NUM))))
      in_push(t, make_error(t, E_TYPE));
    else {
      int i, j, k;

      k = (name->type == V_STRING) ? strlen(name->string) : name->num;
      i = (lidx && (lidx->type == V_NUM)) ? lidx->num : 1;
      j = (ridx && (ridx->type == V_NUM)) ? ridx->num : k;
      in_push(t, extract_elements(t, name, i, j));
    }
    var_free(name);
    var_free(ridx);
    var_free(lidx);
  }
  t->current = t->current->next;
}

/*
 * take the value on top of the stack, change it into a single element list,
 * and push this list back on the stack. This and combine_lists are mainly
 * used to build lists on the stack.
 */
void ntrp_single_elt_list(t)
     task *t;
{
  var *elt, *list;

  elt = in_pop(t);
  list = newlist(1);
  list->list[0] = elt;
  in_push(t, list);
  t->current = t->current->next;
}

/*
 * the top two stack elements should be lists which are to be appended.
 * the top stack element is to be appended to the one right after it on the
 * stack. Thus the stack looks like this before this operation:
 *
 * --> end   (end of list)
 *     begin (beginning of list)
 *     ...
 *
 * the resulting merged list is to be pushed back onto the stack
 */
void ntrp_combine_lists(t)
     task *t;
{
  var *list1, *list2, *result;

  list2 = in_pop(t);
  list1 = in_pop(t);
  if ((list1->type != V_LIST) || (list2->type != V_LIST))
    in_push(t, make_error(t, E_TYPE));
  else {
    int x;

    result = newlist(list1->num + list2->num);
    for (x=0; x<list1->num; x++)
      result->list[x] = var_ref(list1->list[x]);
    for (x=0; x<list2->num; x++)
      result->list[x+list1->num] = var_ref(list2->list[x]);
    in_push(t, result);
  }
  var_free(list1);
  var_free(list2);
  t->current = t->current->next;
}

/*
 * we are trying to splice a list into another list.
 * we don't really do anything to the stack, just make sure that the top
 * element is indeed a list
 */
void ntrp_splice_lists(t)
     task *t;
{
  var *list;

  list = in_pop(t);
  if (list->type != V_LIST) {
    var_free(list);
    in_push(t, make_error(t, E_TYPE));
  } else
    in_push(t, list);
  t->current = t->current->next;
}

/*
 * create an empty (zero length) list and push it on the stack
 */
void ntrp_empty_list(t)
     task *t;
{
  in_push(t, newlist(0));
  t->current = t->current->next;
}

/*
 * push a token onto the stack as a sentinel for end_assign.
 * end_assign takes care of the actual assignment
 */
void ntrp_local_assign(t)
     task *t;
{
  in_push(t, newtoken(MN_LOCAL_ASSIGN));
  t->current = t->current->next;
}

/*
 * push a token onto the stack as a sentinel for end_assign.
 * end_assign takes care of the actual assignment
 */
void ntrp_global_assign(t)
     task *t;
{
  in_push(t, newtoken(MN_GLOBAL_ASSIGN));
  t->current = t->current->next;
}

void ntrp_end_assign(t)
     task *t;
{
  var *v1, *v2, *token;
  Error assign_err;
  int x;

  assign_err = E_NONE;
  v1 = in_pop(t); /* rhs of assignment */
  token = in_pop(t); /* token telling what to do next */
  while ((token->num != MN_GLOBAL_ASSIGN) && (token->num != MN_LOCAL_ASSIGN)) {
    if (token->num == MN_ELEMENT) {
      var *idx, *lhs, *rhs;

      idx = in_pop(t);
      lhs = in_pop(t);
      if (idx->type != V_NUM)
	assign_err = E_TYPE;
      if (assign_err == E_NONE)
	switch (lhs->type) {
	case V_STRING:
	  if (strlen(v1->string) != 1)
	    assign_err = E_INVARG;
	  else if ((idx->num > strlen(lhs->string)) || (idx->num < 1))
	    assign_err = E_RANGE;
	  else {
	    rhs = var_dup(lhs);
	    rhs->string[idx->num-1] = v1->string[0];
	    var_free(v1);
	    v1 = rhs;
	  }
	  break;
	case V_LIST:
	  if ((idx->num > lhs->num) || (idx->num < 1))
	    assign_err = E_RANGE;
	  else {
	    rhs = newlist(lhs->num);
	    for (x=0; x < rhs->num; x++)
	      rhs->list[x] = var_ref(lhs->list[x]);
	    var_free(rhs->list[idx->num-1]);
	    rhs->list[idx->num-1] = v1;
	    v1 = rhs;
	  }
	  break;
	default:
	  assign_err = E_TYPE;
	  break;
	}
      var_free(idx);
      var_free(lhs);
    } else if (token->num == MN_ELEMENT_RANGE) {
      var *i1, *i2, *lhs, *rhs;

      i2 = in_pop(t);
      i1 = in_pop(t);
      lhs = in_pop(t);
      if ((i1->type != V_NUM) || (i2->type != V_NUM))
	assign_err = E_TYPE;
      if (i2->num < i1->num)
	assign_err = E_INVARG;
      if (assign_err == E_NONE)
	switch (lhs->type) {
	case V_STRING:
	  if ((strlen(v1->string) != (i2->num - i1->num +1)) ||
	      (i2->num < i1->num))
	    assign_err = E_INVARG;
	  else if ((i2->num > strlen(lhs->string)) || (i1->num < 1))
	    assign_err = E_RANGE;
	  else {
	    rhs = var_dup(lhs);
	    for (x=i1->num; x<=i2->num; x++)
	      rhs->string[x-1] = v1->string[x-i1->num];
	    var_free(v1);
	    v1 = rhs;
	  }
	  break;
	case V_LIST:
	  if ((i2->num > lhs->num) || (i1->num < 1))
	    assign_err = E_RANGE;
	  else {
	    rhs = newlist(lhs->num);
	    for (x=0; x<rhs->num; x++)
	      rhs->list[x] = var_ref(lhs->list[x]);
	    for (x=i1->num; x<=i2->num; x++) {
	      var_free(rhs->list[x-1]);
	      rhs->list[x-1] = var_ref(v1->list[x-i1->num]);
	    }
	    var_free(v1);
	    v1 = rhs;
	  }
	  break;
	default:
	  assign_err = E_TYPE;
	  break;
	}
      var_free(i1);
      var_free(i2);
      var_free(lhs);
    } else {
      v2 = in_pop(t);
      assign_err = v2->err;
      var_free(v2);
    }
    var_free(token);
    token = in_pop(t);
  }
  if (assign_err != E_NONE) {
    var_free(v1);
    v1 = make_error(t, assign_err);
  }
  v2 = in_pop(t); /* name of the variable to assign to */

  if (token->num == MN_LOCAL_ASSIGN)
    t->local = set_var(t->local, v2->string, v1);
  else if (token->num == MN_GLOBAL_ASSIGN)
    t->buf->global = set_var(t->buf->global, v2->string, v1);
  var_free(token);
  var_free(v2);
  in_push(t, v1);
  t->current = t->current->next;
}

/*
 * the top of the stack will look like this:
 *
 * --> arguments
 *     function name
 *     ...
 *
 * pop these off the stack and execute the function
 * the function *must* push one and only one value back onto this task's
 * stack: that function's return value
 */
void ntrp_function(t)
     task *t;
{
  var *args, *func;
  program *p;

  args = in_pop(t);
  func = in_pop(t);
  if ((func->type != V_STRING) || (args->type != V_LIST))
    in_push(t, make_error(t, E_TYPE));
  else {
    switch (func_exists(global_funcs, func->string)) {
    case 0:
      in_push(t, make_error(t, E_VERBNF));
      break;
    case FUNC_BUILTIN:
      exec_builtin(global_funcs, func->string, t, args);
      break;
    case FUNC_STALKCODE:
      p = find_program(global_funcs, func->string);
      exec_task(t, p, args, 0);
    }
  }
  var_free(args);
  var_free(func);
  t->current = t->current->next;
}

/*
 * the condition's result should be on the stack.
 * if true, continue with with first result.
 * if false, jump to the second result.
 */
void ntrp_conditional(t)
     task *t;
{
  var *cond;

  cond = in_pop(t);
  if (!istrue(cond))
    t->current = t->current->jump;
  else
    t->current = t->current->next;
  var_free(cond);
}

/*
 * task the top two elements off the stack and push true if they're
 * equivalent, false if not
 */
void ntrp_equal(t)
     task *t;
{
  var *rhs, *lhs;

  rhs = in_pop(t);
  lhs = in_pop(t);
  in_push(t, newnum(equal_vars(lhs, rhs)));
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * task the top two elements off the stack and push false if they're
 * equivalent, true if not
 */
void ntrp_not_equal(t)
     task *t;
{
  var *rhs, *lhs;

  rhs = in_pop(t);
  lhs = in_pop(t);
  in_push(t, newnum(!equal_vars(lhs, rhs)));
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs >= rhs back on
 */
void ntrp_greater_equal(t)
     task *t;
{
  var *rhs, *lhs;
  int x;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != rhs->type) ||
      (lhs->type != V_NUM && lhs->type != V_STRING && lhs->type != V_ERR))
    in_push(t, make_error(t, E_TYPE));
  else {
    if (lhs->type == V_NUM)
      x = lhs->num - rhs->num;
    else if (lhs->type == V_STRING)
      x = mystrcasecmp(lhs->string, rhs->string);
    else
      x = lhs->err - rhs->err;
    in_push(t, newnum(x >= 0));
  }
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs > rhs back on
 */
void ntrp_greater_than(t)
     task *t;
{
  var *rhs, *lhs;
  int x;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != rhs->type) ||
      (lhs->type != V_NUM && lhs->type != V_STRING && lhs->type != V_ERR))
    in_push(t, make_error(t, E_TYPE));
  else {
    if (lhs->type == V_NUM)
      x = lhs->num - rhs->num;
    else if (lhs->type == V_STRING)
      x = mystrcasecmp(lhs->string, rhs->string);
    else
      x = lhs->err - rhs->err;
    in_push(t, newnum(x > 0));
  }
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs <= rhs back on
 */
void ntrp_less_equal(t)
     task *t;
{
  var *rhs, *lhs;
  int x;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != rhs->type) ||
      (lhs->type != V_NUM && lhs->type != V_STRING && lhs->type != V_ERR))
    in_push(t, make_error(t, E_TYPE));
  else {
    if (lhs->type == V_NUM)
      x = lhs->num - rhs->num;
    else if (lhs->type == V_STRING)
      x = mystrcasecmp(lhs->string, rhs->string);
    else
      x = lhs->err - rhs->err;
    in_push(t, newnum(x <= 0));
  }
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs < rhs back on
 */
void ntrp_less_than(t)
     task *t;
{
  var *rhs, *lhs;
  int x;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != rhs->type) ||
      (lhs->type != V_NUM && lhs->type != V_STRING && lhs->type != V_ERR))
    in_push(t, make_error(t, E_TYPE));
  else {
    if (lhs->type == V_NUM)
      x = lhs->num - rhs->num;
    else if (lhs->type == V_STRING)
      x = mystrcasecmp(lhs->string, rhs->string);
    else
      x = lhs->err - rhs->err;
    in_push(t, newnum(x < 0));
  }
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs + rhs back on
 */
void ntrp_add(t)
     task *t;
{
  var *lhs, *rhs, *result;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type == V_NUM) && (rhs->type == V_NUM))
    result = newnum(lhs->num + rhs->num);
  else if ((lhs->type == V_STRING) && (rhs->type == V_STRING)) {
    result = newstring((char *)malloc(strlen(lhs->string)
				      +strlen(rhs->string)+1));
    strcpy(result->string, lhs->string);
    strcat(result->string, rhs->string);
  } else
    result = make_error(t, E_TYPE);
  var_free(lhs);
  var_free(rhs);
  in_push(t, result);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs - rhs back on
 */
void ntrp_sub(t)
     task *t;
{
  var *lhs, *rhs;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != V_NUM) || (rhs->type != V_NUM))
    in_push(t, make_error(t, E_TYPE));
  else
    in_push(t, newnum(lhs->num - rhs->num));
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs * rhs back on
 */
void ntrp_mul(t)
     task *t;
{
  var *lhs, *rhs;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != V_NUM) || (rhs->type != V_NUM))
    in_push(t, make_error(t, E_TYPE));
  else
    in_push(t, newnum(lhs->num * rhs->num));
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs / rhs back on
 */
void ntrp_div(t)
     task *t;
{
  var *lhs, *rhs;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != V_NUM) || (rhs->type != V_NUM))
    in_push(t, make_error(t, E_TYPE));
  else if (rhs->num == 0)
    in_push(t, make_error(t, E_DIV));
  else
    in_push(t, newnum(lhs->num / rhs->num));
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * the stack will look like this:
 * --> rhs (right-hand-side)
 *     lhs (left-hand-side)
 *     ...
 *
 * pop these values off the stack and push the result of lhs % rhs back on
 */
void ntrp_mod(t)
     task *t;
{
  var *lhs, *rhs;

  rhs = in_pop(t);
  lhs = in_pop(t);
  if ((lhs->type != V_NUM) || (rhs->type != V_NUM))
    in_push(t, make_error(t, E_TYPE));
  else if (rhs->num == 0)
    in_push(t, make_error(t, E_DIV));
  else
    in_push(t, newnum(lhs->num % rhs->num));
  var_free(lhs);
  var_free(rhs);
  t->current = t->current->next;
}

/*
 * pop the top off the stack and push the negative of this value back on
 */
void ntrp_negate(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  if (v->type != V_NUM)
    in_push(t, make_error(t, E_TYPE));
  else
    in_push(t, newnum(-v->num));
  var_free(v);
  t->current = t->current->next;
}

/*
 * pop the top off the stack
 * push !top back on.
 */
void ntrp_not(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  if (v->type != V_NUM)
    in_push(t, make_error(t, E_TYPE));
  else
    in_push(t, newnum(!v->num));
  var_free(v);
  t->current = t->current->next;
}

/*
 * the second part of the or should be on top of the stack. the short-circuit
 * evaluation should have already snagged and processed the first part.
 * the second part off the stack and push it's truth value back on
 */
void ntrp_or(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  in_push(t, newnum(istrue(v)));
  var_free(v);
  t->current = t->current->next;
}

/*
 * the second part of the and should be on top of the stack. the short-circuit
 * evaluation should have already snagged and processed the first part.
 * the second part off the stack and push it's truth value back on
 */
void ntrp_and(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  in_push(t, newnum(istrue(v)));
  var_free(v);
  t->current = t->current->next;
}

/*
 * perform short-circuit evaluation of the first part of the or. the result
 * of the first condition should be on top of the stack. if it's true,
 * the or is true and we should jump past the second condition. if false,
 * continue with the second condition.
 */
void ntrp_or_ss(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  if (istrue(v)) {
    in_push(t, newnum(1));
    t->current = t->current->jump;
  }
  var_free(v);
  t->current = t->current->next;
}

/*
 * perform short-circuit evaluation of the first part of the and. the result
 * of the first condition should be on top of the stack. if it's false,
 * the and is false and we should jump past the second condition. if true,
 * continue with the second condition.
 */
void ntrp_and_ss(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  if (!istrue(v)) {
    in_push(t, newnum(0));
    t->current = t->current->jump;
  }
  var_free(v);
  t->current = t->current->next;
}

/*
 * the top of the stack should look like this:
 *
 * --> list
 *     item
 *
 * pop these off the stack.
 * if item appears in list, push the index of it's first occurance in list
 * onto the stack, otherwise push 0 (zero) onto the stack
 */
void ntrp_in(t)
     task *t;
{
  var *list, *item, *result;
  int x;

  list = in_pop(t);
  item = in_pop(t);
  if (list->type != V_LIST)
    in_push(t, make_error(t, E_TYPE));
  else {
    result = newnum(0);
    for (x=0; !result->num && x<list->num; x++)
      if (equal_vars(item, list->list[x]))
	result->num = x+1;
    in_push(t, result);
  }
  var_free(list);
  var_free(item);
  t->current = t->current->next;
}

/*
 * we're about to start executing a new line (line in terms of the unparsed
 * function source). save a refernce to this in case the debugger needs it.
 */
void ntrp_line_info(t)
     task *t;
{
  t->line_info = t->current->value;
  t->current = t->current->next;
}

/*
 * we're about to get a list of error that we want to specify debugging
 * behavior on. push a sentinel token onto the stack and turn debugging
 * off until we're done enumerating the errors we want to change.
 */
void ntrp_debug(t)
     task *t;
{
  in_push(t, newtoken(MN_DEBUG));
  t->ignore_errors = 1;
  t->current = t->current->next;
}

/*
 * we're done enumerating the errors we want to specify the behavior for.
 * turn debugging back on.
 */
void ntrp_end_debug(t)
     task *t;
{
  t->ignore_errors = 0;
  t->current = t->current->next;
}

/*
 * assign a function to be called by a set of error conditions.
 * the top of the stack should look like this:
 *
 * --> func_name
 *     error1
 *     ...
 *     errorn
 *     DEBUG      (sentinel placed on the stack to show where the end of
 *                 the error conditions is)
 *     ...
 */
void ntrp_debug_call(t)
     task *t;
{
  var *name, *e;
  char *debugging_func;

  /*
   * pop the function name off the stack
   */
  name = in_pop(t);

  if (t->type != V_STRING) {
    debugging_func = NULL;
    raise_error(t, E_TYPE);
  } else
    debugging_func = mystrdup(name->string);

  var_free(name);

  e = in_pop(t);
  while (e->type == V_ERR) {
    if ((e->err != E_TICKS) && (e->err != E_MAXREC) && debugging_func) {
      t->debug[(int) e->err] = 1;

      if (t->debug_func[(int) e->err])
	free(t->debug_func[(int) e->err]);
      t->debug_func[(int) e->err] = mystrdup(debugging_func);
    }
    
    var_free(e);
    e = in_pop(t);
  }

  var_free(e);
  if (debugging_func)
    free(debugging_func);
  t->current = t->current->next;
}

/*
 * tell the debugger to ignore a set of errors
 * the top of the stack should look like this:
 *
 * --> error1
 *     ...
 *     errorn
 *     DEBUG      (sentinel placed on the stack to show where the end of
 *                 the error conditions is)
 *     ...
 */
void ntrp_debug_ignore(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  while (v->type == V_ERR) {
    if ((v->err != E_TICKS) && (v->err != E_MAXREC)) {
      t->debug[(int) v->err] = 0;

      if (t->debug_func[(int) v->err])
	free(t->debug_func[(int) v->err]);
      t->debug_func[(int) v->err] = NULL;
    }

    var_free(v);
    v = in_pop(t);
  }

  var_free(v);
  t->current = t->current->next;
}

/*
 * tell the debugger to make a set of errors fatal
 * the top of the stack should look like this:
 *
 * --> error1
 *     ...
 *     errorn
 *     DEBUG      (sentinel placed on the stack to show where the end of
 *                 the error conditions is)
 *     ...
 */
void ntrp_debug_fatal(t)
     task *t;
{
  var *v;

  v = in_pop(t);
  while (v->type == V_ERR) {
    if ((v->err != E_TICKS) && (v->err != E_MAXREC)) {
      t->debug[(int) v->err] = 1;

      if (t->debug_func[(int) v->err])
	free(t->debug_func[(int) v->err]);
      t->debug_func[(int) v->err] = NULL;
    }

    var_free(v);
    v = in_pop(t);
  }

  var_free(v);
  t->current = t->current->next;
}
