/*
 * shockit.c v2.2
 * Copyright (c) 1993 Mark Friedman and Gerald Weiler
 * with help from John Wile and Sam Neely
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "shockit.h"

#include <sys/types.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <limits.h>

#ifndef NeXT
# include <unistd.h>
#endif

#ifdef __STDC__
# include <stdarg.h>
# define VA_START(args, form) va_start(args, form)
#else
# include <varargs.h>
# define VA_START(args, form) va_start(args)
#endif

enum sock_type {sock_serv, sock_clnt};
enum sock_family {sock_inet, sock_unix};

#ifndef INADDR_NONE
# define INADDR_NONE ((unsigned long) -1)
#endif

#ifndef MAXHOSTNAMELEN
# define MAXHOSTNAMELEN 64
#endif /* MAXHOSTNAMELEN */

#ifdef _POSIX_SOURCE
# define SOCK_PATHNAME_LEN _POSIX_PATH_MAX
#else
# ifndef MAXPATHLEN
#  define MAXPATHLEN 1024
# endif /* MAXPATHLEN */
# define SOCK_PATHNAME_LEN MAXPATHLEN
#endif /* _POSIX_SOURCE */

#if MAXHOSTNAMELEN > SOCK_PATHNAME_LEN
# define SOCK_NAME_LEN MAXHOSTNAMELEN
#else
# define SOCK_NAME_LEN SOCK_PATHNAME_LEN
#endif

static unsigned long bitmask[32] = {
  0x00000001, 0x00000002, 0x00000004, 0x00000008,
  0x00000010, 0x00000020, 0x00000040, 0x00000080,
  0x00000100, 0x00000200, 0x00000400, 0x00000800,
  0x00001000, 0x00002000, 0x00004000, 0x00008000,
  0x00010000, 0x00020000, 0x00040000, 0x00080000,
  0x00100000, 0x00200000, 0x00400000, 0x00800000,
  0x01000000, 0x02000000, 0x04000000, 0x08000000,
  0x10000000, 0x20000000, 0x40000000, 0x80000000
};

typedef struct {
  enum sock_type type;
  enum sock_family family;

  int socket_fd;
  FILE *socket_f;

  /* this has to be able to hold a pathname or a hostname */
  char name[SOCK_NAME_LEN];

  union {
    struct sockaddr_un uaddr;
    struct sockaddr_in iaddr;
  } addr;

  /* any machine supporting TCP/IP should have at least 32 bit longs */
  unsigned long strip_chars;
  unsigned long eol_chars;

  int connected;
} iSocket; /* internal Socket type */

#ifdef NEED_INT_DEFINES
int socket();
int bind();
int listen();
int setsockopt();
int getsockname();
int gethostname();
int connect();
int bzero();
int select();
int accept();
int vfprintf();
int fflush();
int fclose();
#endif

int sock_make_server (server, path, port)
     Socket *server;
     char *path;
     int port;

/*
 * requires: none
 * ensures:  server is created and points to a server socket
 *           listening on the requested port or path if the call is
 *           successful. server points to NULL on failure.
 *           if port == SOCK_UNIXDOM then a unix domain socket server is
 *              created with the path 'path'.
 *           else an inet socket server is created on the port 'port' and the
 *              'path' parameter is ignored.
 * returns:  0 on success, -1 on failure
 * errors:   returns any error condition detected by socket(), bind(), or
 *           getsockname(), or EINVAL if an invalid port number was specified.
 */

{
  iSocket *nuserv;
  int iaddr_len = sizeof(struct sockaddr_in);
  struct hostent *hent;
  char buf[MAXHOSTNAMELEN]; /* MAXHOSTNAMELEN from sys/param.h */
  int temp = 1;

  *server = NULL;

  nuserv = (iSocket *)malloc(sizeof(iSocket));
  nuserv->type = sock_serv;
  if (port == SOCK_UNIXDOM)
    nuserv->family = sock_unix;
  else if (port >= 0)
    nuserv->family = sock_inet;
  else {
    errno = EINVAL;
    return -1;
  }

  /*
   * need this for correct behavior on select() and write()
   * repeated for every socket, but makes sure it's done! :-)
   */
  signal(SIGPIPE,SIG_IGN);

  switch (nuserv->family) {
  case sock_unix:
    /*
     * create an endpoint for communication
     */
    if((nuserv->socket_fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
      /* Ack! no more file descriptors left! */
      free(nuserv);
      return -1;
    }

    /*
     * use UNIX domain
     */
    nuserv->addr.uaddr.sun_family = AF_UNIX;

    /*
     * bind a name to the socket
     */
    strcpy(nuserv->addr.uaddr.sun_path, path);
    if(bind(nuserv->socket_fd, (void *) &nuserv->addr.uaddr,
	    sizeof(nuserv->addr.uaddr.sun_family) + strlen(path)) < 0) {
      close(nuserv->socket_fd);
      free(nuserv);
      return -1;
    }

    /*
     * queue 5 sockets if busy
     */
    listen(nuserv->socket_fd, 5);

    /*
     * store pathname in structure
     */
    strcpy(nuserv->name, path);

    break;
  case sock_inet:
    /*
     * create an endpoint for communication
     */
    if((nuserv->socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      /* Ack! no more file descriptors left! */
      free(nuserv);
      return -1;
    }

    /*
     * toggle local address reuse
     * some systems aren't good at clean-up
     */
    setsockopt(nuserv->socket_fd, SOL_SOCKET, SO_REUSEADDR,
	       (char *)&temp, sizeof(temp));

    /*
     * use Internet puppies
     */
    nuserv->addr.iaddr.sin_family = AF_INET;

    /*
     * not choosy about who calls
     */
    nuserv->addr.iaddr.sin_addr.s_addr = INADDR_ANY;

    /*
     * the port the server will listen on
     */
    nuserv->addr.iaddr.sin_port = htons(port);

    /*
     * bind a name to the socket
     */
    if(bind(nuserv->socket_fd, (void *) &nuserv->addr.iaddr,
	    sizeof(struct sockaddr_in)) < 0) {
      close(nuserv->socket_fd);
      free(nuserv);
      return -1;
    }

    /*
     * get socket name
     */
    if(getsockname(nuserv->socket_fd, (void *) &nuserv->addr.iaddr,
		   &iaddr_len) < 0) {
      close(nuserv->socket_fd);
      free(nuserv);
      return -1;
    }

    /*
     * queue 5 sockets if busy
     */
    listen(nuserv->socket_fd, 5);

    /*
     * get (short) name of current host
     */
    gethostname(buf, MAXHOSTNAMELEN);

    /*
     * get all info of the host (including full name)
     */
    hent = gethostbyname(buf);

    /*
     * store full name in structure
     */
    strcpy(nuserv->name, hent->h_name);

    break;
  default:
    /* this will never happen :-/ */
    close(nuserv->socket_fd);
    free(nuserv);
    errno = EINVAL;
    return -1;
    break;
  }

  /*
   * initialize to 'connected'
   */
  nuserv->connected = 1;

#ifdef SOCK_DEBUG
  /*
   * announce establishment
   */
  printf("#### We are %s, port %d (desc %d) ####\n",
         nuserv->name, ntohs(nuserv->addr.iaddr.sin_port), nuserv->socket_fd);
#endif /* SOCK_DEBUG */

  *server = (Socket)nuserv;

  return 0;
}

int sock_call_server (connection, name, port)
     Socket *connection;
     char *name;
     int port;

/*
 * requires: none
 * ensures:  connection is created and points to a client socket
 *           which has a connection to the requested host/port pair or path
 *           if the call was successful. connection points to NULL on failure.
 *           if port == SOCK_UNIXDOM then it will try to attach to a unix
 *              domain server on the path 'name'.
 *           else it will try to attach to an inet socket on host 'name' and
 *              port 'port'.
 * returns:  0 on success, -1 on failure
 * errors:   returns any error condition detected by socket(), connect(),
 *           gethostbyname(), or EINVAL if an invalid port number was given.
 */

{
  iSocket *nucon;
  struct hostent *remote_host; /* internet numbers, names */

  *connection = NULL;
  nucon = (iSocket *)malloc(sizeof(iSocket));
  nucon->type = sock_clnt;
  if (port == SOCK_UNIXDOM)
    nucon->family = sock_unix;
  else if (port >= 0)
    nucon->family = sock_inet;
  else {
    errno = EINVAL;
    return -1;
  }

  /*
   * need this for correct behavior on select() and write()
   * repeated for every socket, but makes sure it's done! :-)
   */
  signal(SIGPIPE,SIG_IGN);

  switch (nucon->family) {
  case sock_unix:
    /*
     * create an endpoint for communication
     */
    if((nucon->socket_fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
      /* Ack! no more file descriptors left! */
      free(nucon);
      return -1;
    }

    /*
     * use UNIX domain
     */
    nucon->addr.uaddr.sun_family = AF_UNIX;

    /*
     * target the socket we're calling
     */
    strcpy(nucon->addr.uaddr.sun_path, name);

    /*
     * try to hook up with the server
     */
    if(connect(nucon->socket_fd, (struct sockaddr *) &nucon->addr.uaddr,
	       sizeof(nucon->addr.uaddr.sun_family)+strlen(name)) < 0) {
      /* tell the folx nobody's listening at that path number */
      close(nucon->socket_fd);
      free(nucon);
      return -1;
    }

    /* record the pathname */
    strcpy(nucon->name, name);

    break;
  case sock_inet:
    /*
     * create an endpoint for communication
     */
    if((nucon->socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      /* Ack! no more file descriptors left! */
      free(nucon);
      return -1;
    }

    /*
     * use Internet puppies
     */
    nucon->addr.iaddr.sin_family = AF_INET;

    /*
     * get the address of the target machine
     */

    /* check if we have a numeric quad */
    nucon->addr.iaddr.sin_addr.s_addr = inet_addr(name);
    if (nucon->addr.iaddr.sin_addr.s_addr == INADDR_NONE) {
      /* name is a hostname */
      if ((remote_host = gethostbyname(name)) == NULL) {
	/* tell the folx we couldn't find the host machine */
	close(nucon->socket_fd);
	free(nucon);
	return -1;
      } else
	(void) memcpy((char *) &nucon->addr.iaddr.sin_addr,
		      (char *) remote_host->h_addr, remote_host->h_length);
    } else
      remote_host = gethostbyaddr((char *) &nucon->addr.iaddr.sin_addr,
				  sizeof(struct in_addr), AF_INET);

    /*
     * target the port we're calling
     */
    nucon->addr.iaddr.sin_port = htons(port);

    /* try to hook up with the server */
    if(connect(nucon->socket_fd, (struct sockaddr *) &nucon->addr.iaddr,
	       sizeof(nucon->addr.iaddr)) < 0) {
      /* tell the folx nobody's listening at that port number */
      close(nucon->socket_fd);
      free(nucon);
      return -1;
    }

    /* record the host address */
    if (remote_host)
      strcpy(nucon->name, remote_host->h_name);
    else
      strcpy(nucon->name, inet_ntoa(nucon->addr.iaddr.sin_addr));

    break;
  default:
    /* this will never happen :-/ */
    close(nucon->socket_fd);
    free(nucon);
    errno = EINVAL;
    return -1;
    break;
  }

  /* the socket_f part is used by sock_printf */
  if ((nucon->socket_f = fdopen(nucon->socket_fd, "w")) == NULL) {
    close(nucon->socket_fd);
    free(nucon);
    return -1;
  }

  nucon->connected = 1;
  nucon->eol_chars = bitmask['\n'];
  nucon->strip_chars = bitmask['\r'];
  *connection = (Socket)nucon;

#ifdef SOCK_DEBUG
  /* announce connection */
  printf("#### We called %s, port %d (desc %d) ####\n",
         nucon->name, ntohs(nucon->addr.iaddr.sin_port), nucon->socket_fd);
#endif /* SOCK_DEBUG */
  
  return 0;
}

int sock_accept (server, connection, timeout)
     Socket server;
     Socket *connection;
     int timeout;

/*
 * requires: server has been opened sucessfully with make_server
 * ensures:  listens for timeout milliseconds (or indefinitely if timeout
 *           is zero.) if a caller is found within the timeout period,
 *           connection is created and points to a connection to the
 *           caller. if no listener is found within the timeout period,
 *           connection points to NULL. 0 is returned in either of these
 *           cases, -1 is returned on abnormal failure.
 * returns:  0 on success, -1 on failure
 * errors:   returns any error condition detected by select() or accept(),
 *           plus the following error conditions:
 *           EINVAL      server is not an open socket
 *           EOPNOTSUPP  server was not opened with sock_make_server
 *                       (server is a client socket)
 */

{
  fd_set readfds;
  struct timeval Timeout;
  struct hostent *hent;
  int uaddr_len = sizeof(struct sockaddr_un);
  int iaddr_len = sizeof(struct sockaddr_in);
  iSocket *nucon, *serv;

  *connection = NULL;
  serv = (iSocket *) server;

  /*
   * see if serv is actually a server
   */
  if (!serv) {
    errno = EINVAL;
    return -1;
  }

  if (serv->type != sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  /*
   * get the time we wait for new connect
   */
  if (timeout) {
    Timeout.tv_usec  = (timeout % 1000) * 1000;
    Timeout.tv_sec = timeout / 1000;
  }

  /*
   * reset file descriptors and mark the one we want to test (serv)
   */
  FD_ZERO(&readfds);
  FD_SET(serv->socket_fd, &readfds);

  /*
   * check if someone wants to connect
   */
  if(select(FD_SETSIZE, &readfds, NULL, NULL, timeout ? &Timeout : NULL) < 0)
    return -1;

  /*
   * if nobody on the line, return zero & leave connection NULL
   */
  if(!FD_ISSET(serv->socket_fd, &readfds))
    return 0;

  /*
   * initialize a socket
   */
  nucon = (iSocket *)malloc(sizeof(iSocket));
  nucon->type = sock_clnt;
  nucon->family = serv->family;

  switch (nucon->family) {
  case sock_inet:
    /* accept a connection on this_socket */
    if((nucon->socket_fd = accept(serv->socket_fd, (void *) &nucon->addr.iaddr,
				  &iaddr_len)) < 0) {
      free(nucon);
      return -1;
    }

    /* get all info for the caller */
    if((hent = gethostbyaddr((char *)&nucon->addr.iaddr.sin_addr,
			     sizeof(struct in_addr), AF_INET)) == NULL)
      /* he's refusing us! */
      strcpy(nucon->name, inet_ntoa(nucon->addr.iaddr.sin_addr));
    else
      /* record the caller's address */
      strcpy(nucon->name, hent->h_name);

    break;
  case sock_unix:
    /* accept a connection on this_socket */
    if((nucon->socket_fd = accept(serv->socket_fd, (void *) &nucon->addr.uaddr,
				  &uaddr_len)) < 0) {
      free(nucon);
      return -1;
    }

    strcpy(nucon->name, serv->name);
    break;
  default:
    /* this will never happen :-/ */
    close(nucon->socket_fd);
    free(nucon);
    errno = EINVAL;
    return -1;
    break;
  }

  /* the socket_f part is used by sock_printf */
  nucon->socket_f = fdopen(nucon->socket_fd, "w");

  /* initialize to 'connected' */
  nucon->connected = 1;
  nucon->eol_chars = bitmask['\n'];
  nucon->strip_chars = bitmask['\r'];
  *connection = (Socket)nucon;

#ifdef SOCK_DEBUG
  /* announce connection */
  printf(">>>> Opened %s, port %d (desc %d) <<<<\n", nucon->name,
         ntohs(nucon->addr.iaddr.sin_port), nucon->socket_fd);
#endif /* SOCK_DEBUG */

  /* return success */
  return 0;
}

int sock_strip_char(sock, c)
     Socket sock;
     int c;

/*
 * requires: sock is an open socket and is not a server
 * ensures:  if 0<=c<=0x1f, then the character will be discarded by sock_gets
 *           if c == SOCK_STRIP_NONE, then sock will be set to a state of not
 *           discarding anything.
 *           if c == SOCK_STRIP_DEFAULT, then sock will be set back to the
 *           default (which is to strip '\r' characters only.)
 * returns:  0 on success, -1 on failure
 * errors:   EINVAL       sock not open or c is out of range
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->type == sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  if (c == SOCK_STRIP_NONE)
    isock->strip_chars = 0;
  else if (c == SOCK_STRIP_DEFAULT)
    isock->strip_chars = bitmask['\r'];
  else if (c & ~0x1f) {
    errno = EINVAL;
    return -1;
  } else
    isock->strip_chars |= (bitmask[c]);

  return 0;
}

int sock_eol_char(sock, c)
     Socket sock;
     int c;

/*
 * requires: sock is an open socket and is not a server
 * ensures:  if 0<=c<=0x1f, then the character will be treated as an EOL
 *           by sock_gets.
 *           if c == SOCK_EOL_NONE, then sock will be set to a state of not
 *           treating anything as EOL.
 *           if c == SOCK_EOL_DEFAULT, then sock will be set back to the
 *           default (which is to treat '\n' as EOL only.)
 * returns:  0 on success, -1 on failure
 * errors:   EINVAL       sock not open or c is out of range
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->type == sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  if (c == SOCK_EOL_NONE)
    isock->eol_chars = 0;
  else if (c == SOCK_EOL_DEFAULT)
    isock->eol_chars = bitmask['\n'];
  else if (c & ~0x1f) {
    errno = EINVAL;
    return -1;
  } else
    isock->eol_chars |= (bitmask[c]);

  return 0;
}

int sock_gets (sock, buff, nchars, timeout)
     Socket sock;
     char *buff;
     int nchars;
     int timeout;

/*
 * requires: sock has been opened successfully and is not a server socket
 *           buff has at least nchars characters allocated and nchars > 0
 * ensures:  will read chars until nchars-1 characters have been read, OR
 *           until an EOL has been read, OR
 *           until timeout has expired (if timeout > 0)
 *           timeout is given in milliseconds
 *           buff will always be null terminated.
 * returns:  number of characters read, or -1 on failure
 * errors:   any error returned by select() or read() plus the following:
 *           EINVAL       sock not open or nchars <= 0
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

{
  struct timeval Timeout;
  int numread, sel_ret;
  iSocket *isock;
  char c;
  fd_set readfds;

  if (nchars < 1) {
    errno = EINVAL;
    return -1;
  }

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->type == sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  if (timeout) {
    Timeout.tv_usec  = (timeout % 1000) * 1000;
    Timeout.tv_sec = timeout / 1000;
  }

  numread = 0;

  while (numread < nchars-1) {
    FD_ZERO(&readfds);
    FD_SET(isock->socket_fd, &readfds);
    if ((sel_ret = select(FD_SETSIZE, &readfds, NULL, NULL,
			  timeout ? &Timeout : NULL)) < 0) {
      isock->connected = 0;
      if (numread)
	return numread;
      else
	return -1;
    }

    if (!FD_ISSET(isock->socket_fd, &readfds))
      break;

    if (read(isock->socket_fd, &c, 1) < 1) {
      isock->connected = 0;
      if (numread)
	return numread;
      else
	return -1;
    }

    if (c & ~0x1f) {
      buff[numread++] = c;
    } else {
      if (!(isock->strip_chars & bitmask[(int) c]))
	buff[numread++] = c;
      if (isock->eol_chars & bitmask[(int) c])
	break;
    }
  }

  if (!numread)
    return -1;

  buff[numread] = '\0';

  return numread;
}

int sock_read (sock, buff, nchars, timeout)
     Socket sock;
     char *buff;
     int nchars;
     int timeout;

/*
 * requires: sock has been opened successfully and is not a server socket
 * ensures:  will reach chars into buff until nchars have been read,
 *              OR until timeout has expired if timeout > 0
 *           timeout is given in milliseconds
 * returns:  number of characters read, or -1 on failure
 * errors:   any error returned by select() or read() plus the following:
 *           EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

{
  struct timeval Timeout;
  int numread, sel_ret;
  iSocket *isock;
  char c;
  fd_set readfds;

  if (nchars < 0)
    return -1;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->type == sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  if (timeout) {
    Timeout.tv_usec  = (timeout % 1000) * 1000;
    Timeout.tv_sec = timeout / 1000;
  }

  numread = 0;

  while (numread < nchars) {
    FD_ZERO(&readfds);
    FD_SET(isock->socket_fd, &readfds);
    if ((sel_ret = select(FD_SETSIZE, &readfds, NULL, NULL,
			  timeout ? &Timeout : NULL)) < 0) {
      isock->connected = 0;
      if (numread)
	return numread;
      else
	return -1;
    }

    if (!FD_ISSET(isock->socket_fd, &readfds))
      break;

    if (read(isock->socket_fd, &c, 1) < 1) {
      isock->connected = 0;
      if (numread)
	return numread;
      else
	return -1;
    }

    buff[numread++] = c;
  }

  return numread;
}

#ifdef __STDC__
int sock_printf (Socket sock, char *form, ...)
#else
int sock_printf (sock, form, va_alist)
     Socket sock;
     char *form;
     va_dcl
#endif /* __STDC__ */

/*
 * requires: sock has been opened sucessfully and is not a server socket
 * ensures:  form is taken as a format argument in the style of printf
 *           and the resulting string is written to the socket.
 * returns:  number of characters written on success, -1 on failure
 * errors:   any error returned by vsprintf() plus the following:
 *           EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

{
  iSocket *isock;
  int write_ret;
  va_list list;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->type == sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  VA_START(list, form);
  if ((write_ret = vfprintf(isock->socket_f, form, list)) == -1)
    isock->connected = 0;
  else
    fflush(isock->socket_f);
  va_end(list);

  return write_ret;
}

int sock_write (sock, buff, nchars)
     Socket sock;
     char *buff;
     int nchars;

/*
 * requires: sock has been opened successfully and is not a server socket
 * ensures:  nchars from buff are written to the socket.
 * returns:  number of chatacters written on success, -1 on failure
 * errors:   any error returned by vsprintf() plus the following:
 *           EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

{
  iSocket *isock;
  int write_ret;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->type == sock_serv) {
    errno = EOPNOTSUPP;
    return -1;
  }

  if ((write_ret = write(isock->socket_fd, buff, nchars)) == -1)
    isock->connected = 0;

  return write_ret;
}

int sock_fd (sock)
     Socket sock;

/*
 * requires: sock is an open Socket
 * ensures:  none
 * returns:  the file descriptor of sock on success, -1 on failure
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  return isock->socket_fd;
}

int sock_select (fds, nfd, ready, timeout)
     int fds[];
     int nfd;
     char ready[];
     int timeout;

/*
 * requires: fds points to an array of nfd open file descriptors
 *           ready is an array of at least nfd characters
 * ensures:  each element of ready will be set to one of the following:
 *           1 if data is ready to be read from the file descriptor
 *           in the corresponding spot in fds.
 *           0 if no data is ready to be read.
 * returns:  0 if none of the sockets are ready to be read after timeout
 *           milliseconds have elapsed, the number ready to be read on success,
 *           -1 on failure.
 * errors:   any error returned by select()
 */

{
  int x, sel_ret;
  fd_set readfds;
  struct timeval Timeout;

  Timeout.tv_usec  = (timeout % 1000) * 1000;
  Timeout.tv_sec = timeout / 1000;

  FD_ZERO(&readfds);
  for (x=0; x<nfd; x++)
    FD_SET(fds[x], &readfds);

  if ((sel_ret = select(FD_SETSIZE, &readfds, NULL, NULL,
			timeout ? &Timeout : NULL)) < 0)
    return -1;

  for (x=0; x<nfd; x++) {
    if (FD_ISSET(fds[x], &readfds))
      ready[x] = 1;
    else
      ready[x] = 0;
  }

  return sel_ret;
}

int sock_connected (sock)
     Socket sock;

/*
 * requires: sock has been opened sucessfully and is not a server socket
 * ensures:  none
 * returns:  1 is sock is connected, 0 if not, -1 on failure
 * errors:   EINVAL       sock not open
 */

{
  if (!sock) {
    errno = EINVAL;
    return -1;
  }

  return ((iSocket *)sock)->connected;
}

int sock_close (sock)
     Socket *sock;

/*
 * requires: sock has been opened sucessfully
 * ensures:  sock is closed
 * returns:  0 on success, -1 on failure
 * errors:   EINVAL       sock not open
 */

{
  iSocket *isock;

  if (!sock || !*sock) {
    errno = EINVAL;
    return -1;
  }

  isock = (iSocket *) *sock;
  fclose(isock->socket_f);
  free(isock);
  *sock = NULL;
  return 0;
}

int sock_family (sock)
     Socket sock;

/*
 * requires: sock has been opened sucessfully
 * ensures:  none
 * returns:  SOCK_UNIX or SOCK_INET on success, -1 on failure
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;
  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  switch (isock->family) {
  case sock_inet:
    return SOCK_INET;
    break;
  case sock_unix:
    return SOCK_UNIX;
    break;
  default:
    return -1;
    break;
  }
}

int sock_host (sock, name)
     Socket sock;
     char *name;

/*
 * requires: sock has been opened sucessfully and is a SOCK_INET family socket.
 *           name is an array that has enough memory to hold a hostname
 * ensures:  name will contain the hostname of the host on the remote end
 *           of the socket if available
 * returns:  0 on success, -1 on failure (hostname not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_INET socket
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->family != sock_inet) {
    errno = EOPNOTSUPP;
    return -1;
  }

  strcpy(name, isock->name);
  return name[0] ? 0 : -1;
}

int sock_ipaddr (sock, ipaddr)
     Socket sock;
     char *ipaddr;

/*
 * requires: sock has been opened sucessfully and is a SOCK_INET family socket.
 *           ipaddr has enough storage allocated to hold an ip address
 *           in d.d.d.d form
 * ensures:  ipaddr will contain the ip address of the remote system
 * returns:  0 on success, -1 on failure (ip address not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_INET socket
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->family != sock_inet) {
    errno = EOPNOTSUPP;
    return -1;
  }

  strcpy(ipaddr, inet_ntoa(isock->addr.iaddr.sin_addr));
  return ipaddr[0] ? 0 : -1;
}

int sock_port (sock, port)
     Socket sock;
     int *port;

/*
 * requires: sock has been opened sucessfully and is a SOCK_INET family socket.
 *           port has been allocated.
 * ensures:  port will contain the port number sock is connected to on the
 *           remote system
 * returns:  0 on success, -1 on failure (port not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_INET socket
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->family != sock_inet) {
    errno = EOPNOTSUPP;
    return -1;
  }

  *port = ntohs(isock->addr.iaddr.sin_port);
  return 0;
}

int sock_path (sock, path)
     Socket sock;
     char *path;

/*
 * requires: sock has been opened sucessfully and is a SOCK_UNIX family socket
 *           or a named pipe.  path is an array that must have at least as
 *           many characters as the maximum pathname length
 * ensures:  path will contain the pathname of the unix domain socket
 *           or named pipe.
 * returns:  0 on success, -1 on failure (path not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_UNIX socket
 */

{
  iSocket *isock;

  isock = (iSocket *)sock;

  if (!isock) {
    errno = EINVAL;
    return -1;
  }

  if (!isock->connected) {
    errno = ENOTCONN;
    return -1;
  }

  if (isock->family != sock_unix) {
    errno = EOPNOTSUPP;
    return -1;
  }

  strcpy(path, isock->name);
  return path[0] ? 0 : -1;
}
