#ifndef _SHOCKIT_H_
#define _SHOCKIT_H_

#ifdef SOCK_P
#undef SOCK_P
#endif
#ifdef __STDC__
#define SOCK_P(x) x
#else
#define SOCK_P(x) ()
#endif

#define SOCK_UNIXDOM       -1

#define SOCK_UNIX           1
#define SOCK_INET           2

#define SOCK_EOL_NONE      -1
#define SOCK_EOL_DEFAULT   -2

#define SOCK_STRIP_NONE    -1
#define SOCK_STRIP_DEFAULT -2

typedef void *Socket;

int sock_make_server SOCK_P((Socket *server, char *path, int port));
/*
 * requires: none
 * ensures:  server is created and points to a server socket
 *           listening on the requested port or path if the call is
 *           successful. server points to NULL on failure.
 *           if port == SOCK_UNIXDOM then a unix domain socket server is
 *              created with the path 'path'.
 *           else an inet socket server is created on the port 'port' and the
 *              'path' parameter is ignored.
 * returns:  0 on success, -1 on failure
 * errors:   returns any error condition detected by socket(), bind(), or
 *           getsockname(), or EINVAL if an invalid port number was specified.
 */

int sock_call_server SOCK_P((Socket *connection, char *name, int port));
/*
 * requires: none
 * ensures:  connection is created and points to a client socket
 *           which has a connection to the requested host/port pair or path
 *           if the call was successful. connection points to NULL on failure.
 *           if port == SOCK_UNIXDOM then it will try to attach to a unix
 *              domain server on the path 'name'.
 *           else it will try to attach to an inet socket on host 'name' and
 *              port 'port'.
 * returns:  0 on success, -1 on failure
 * errors:   returns any error condition detected by socket(), connect(),
 *           gethostbyname(), or EINVAL if an invalid port number was given.
 */

int sock_accept SOCK_P((Socket server, Socket *connection, int timeout));
/*
 * requires: server has been opened sucessfully with make_server
 * ensures:  listens for timeout milliseconds (or indefinitely if timeout
 *           is zero.) if a caller is found within the timeout period,
 *           connection is created and points to a connection to the
 *           caller. if no listener is found within the timeout period,
 *           connection points to NULL. 0 is returned in either of these
 *           cases, -1 is returned on abnormal failure.
 * returns:  0 on success, -1 on failure
 * errors:   returns any error condition detected by select() or accept(),
 *           plus the following error conditions:
 *           EINVAL      server is not an open socket
 *           EOPNOTSUPP  server was not opened with sock_make_server
 *                       (server is a client socket)
 */

int sock_strip_char SOCK_P((Socket sock, int c));
/*
 * requires: sock is an open socket and is not a server
 * ensures:  if 0<=c<=0x1f, then the character will be discarded by sock_gets
 *           if c == SOCK_STRIP_NONE, then sock will be set to a state of not
 *           discarding anything.
 *           if c == SOCK_STRIP_DEFAULT, then sock will be set back to the
 *           default (which is to strip '\r' characters only.)
 * returns:  0 on success, -1 on failure
 * errors:   EINVAL       sock not open or c is out of range
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

int sock_eol_char SOCK_P((Socket sock, int c));
/*
 * requires: sock is an open socket and is not a server
 * ensures:  if 0<=c<=0x1f, then the character will be treated as an EOL
 *           by sock_gets.
 *           if c == SOCK_EOL_NONE, then sock will be set to a state of not
 *           treating anything as EOL.
 *           if c == SOCK_EOL_DEFAULT, then sock will be set back to the
 *           default (which is to treat '\n' as EOL only.)
 * returns:  0 on success, -1 on failure
 * errors:   EINVAL       sock not open or c is out of range
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

int sock_gets SOCK_P((Socket sock, char *buff, int nchars, int timeout));
/*
 * requires: sock has been opened successfully and is not a server socket
 *           buff has at least nchars characters allocated and nchars > 0
 * ensures:  will read chars until nchars-1 characters have been read, OR
 *           until an EOL has been read, OR
 *           until timeout has expired (if timeout > 0)
 *           timeout is given in milliseconds
 *           buff will always be null terminated.
 * returns:  number of characters read, or -1 on failure
 * errors:   any error returned by select() or read() plus the following:
 *           EINVAL       sock not open or nchars <= 0
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

int sock_read SOCK_P((Socket sock, char *buff, int nchars, int timeout));
/*
 * requires: sock has been opened successfully and is not a server socket
 * ensures:  will reach chars into 'buff' until 'nchars' have been read,
 *              OR until timeout has expired if timeout > 0
 *           timeout is given in milliseconds
 * returns:  number of characters read, or -1 on failure
 * errors:   any error returned by select() or read() plus the following:
 *           EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

int sock_printf SOCK_P((Socket sock, char *form, ...));
/*
 * requires: sock has been opened sucessfully and is not a server socket
 * ensures:  form is taken as a format argument in the style of printf
 *           and the resulting string is written to the socket.
 * returns:  number of characters written on success, -1 on failure
 * errors:   any error returned by vsprintf() plus the following:
 *           EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

int sock_write SOCK_P((Socket sock, char *buff, int nchars));
/*
 * requires: sock has been opened successfully and is not a server socket
 * ensures:  nchars from buff are written to the socket.
 * returns:  number of chatacters written on success, -1 on failure
 * errors:   any error returned by write() plus the following:
 *           EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is a server socket
 */

int sock_fd SOCK_P((Socket sock));
/*
 * requires: sock is an open Socket
 * ensures:  none
 * returns:  the file descriptor of sock on success, -1 on failure
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 */

int sock_select SOCK_P((int fds[], int nfd, char ready[], int timeout));
/*
 * requires: fds points to an array of nfd open file descriptors
 *           ready is an array of at least nfd characters
 * ensures:  each element of ready will be set to one of the following:
 *           1 if data is ready to be read from the file descriptor
 *           in the corresponding spot in fds.
 *           0 if no data is ready to be read.
 * returns:  0 if none of the sockets are ready to be read after timeout
 *           milliseconds have elapsed, the number ready to be read on success,
 *           -1 on failure.
 * errors:   any error returned by select()
 */

int sock_connected SOCK_P((Socket sock));
/*
 * requires: sock has been opened sucessfully and is not a server socket
 * ensures:  none
 * returns:  1 is sock is connected, 0 if not, -1 on failure
 * errors:   EINVAL       sock not open
 */

int sock_close SOCK_P((Socket *sock));
/*
 * requires: sock has been opened sucessfully
 * ensures:  sock is closed
 * returns:  0 on success, -1 on failure
 * errors:   EINVAL       sock not open
 */

int sock_family SOCK_P((Socket sock));
/*
 * requires: sock has been opened sucessfully
 * ensures:  none
 * returns:  SOCK_UNIX or SOCK_INET on success, -1 on failure
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 */

int sock_host SOCK_P((Socket sock, char *name));
/*
 * requires: sock has been opened sucessfully and is a SOCK_INET family socket.
 *           name is an array that has enough memory to hold a hostname
 * ensures:  name will contain the hostname of the host on the remote end
 *           of the socket if available
 * returns:  0 on success, -1 on failure (hostname not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_INET socket
 */

int sock_ipaddr SOCK_P((Socket sock, char *ipaddr));
/*
 * requires: sock has been opened sucessfully and is a SOCK_INET family socket.
 *           ipaddr has enough storage allocated to hold an ip address
 *           in d.d.d.d form
 * ensures:  ipaddr will contain the ip address of the remote system
 * returns:  0 on success, -1 on failure (ip address not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_INET socket
 */

int sock_port SOCK_P((Socket sock, int *port));
/*
 * requires: sock has been opened sucessfully and is a SOCK_INET family socket.
 *           port has been allocated.
 * ensures:  port will contain the port number sock is connected to on the
 *           remote system
 * returns:  0 on success, -1 on failure (port not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_INET socket
 */

int sock_path SOCK_P((Socket sock, char *path));
/*
 * requires: sock has been opened sucessfully and is a SOCK_UNIX family socket
 *           or a named pipe.  path is an array that must have at least as
 *           many characters as the maximum pathname length
 * ensures:  path will contain the pathname of the unix domain socket
 *           or named pipe.
 * returns:  0 on success, -1 on failure (path not available)
 * errors:   EINVAL       sock not open
 *           ENOTCONN     sock is no longer connected
 *           EOPNOTSUPP   sock is not a SOCK_UNIX socket
 */

#endif /* _SHOCKIT_H_ */
