/*
 * stalk/stalk.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "stalk.h"
#include "buffer.h"
#include "var.h"
#include "task.h"
#include "stuff.h"
#include "functions.h"
#include "parser.h"
#include "stream.h"
#include "interpret.h"

static char *stalkfile = NULL;
static char init[20][100];
static int initlen;

int done;
int bell_on;
int screen_mode;
int stick_around;
int cols, rows;

int main(argc, argv)
     int argc;
     char **argv;
{
  int x;
  program *p;
  statement *root;
  var *v;

  FILE *fp;

  char *code_buf, *line_buf, tmp[100];
  int code_size, line_size, code_len, line_len;

  char *home;
  int found;
  int tasks_waiting;

  done = 0;
  bell_on = 1;
  screen_mode = TERM_DUMB;
  stick_around = 0;

  srand(time(0));
  parse_args(argc, argv);

  setup_global_functions();

  init_buffers();
  init_vars();
  init_tasks();

  /* setup signal handlers */
#ifndef STICK_AROUND
  signal(SIGSEGV, quit);
#endif
#ifdef SIGBUS
  signal(SIGBUS, quit);
#endif
  signal(SIGHUP, quit);
  signal(SIGILL, quit);
  signal(SIGTERM, quit);
  signal(SIGQUIT, quit);

#ifdef DEBUG_MEM
  /* compile this part in only if you're using dbmalloc */

  setup_memcheck();
#endif /* DEBUG_MEM */

  (setup_screen)();
  (new_buffer)(NULL, "* new buffer *", 0);

  home = getenv("HOME");

  found = 0;
  if (!stalkfile && !(stalkfile = getenv("STALK")) && home) {
    stalkfile = (char *)malloc((strlen(home) + 10) * sizeof(char));
    sprintf(stalkfile, "%s%s", home, "/.stalk");
  }

  if ((stalkfile && (fp = fopen(stalkfile, "r")))
      || (fp = fopen(CONFIG_STALK, "r"))) {
    code_buf = (char *)malloc(code_size = 1000);
    code_buf[0] = '\0';
    code_len = 0;

    line_buf = (char *)malloc(line_size = 100);

    do {
      line_buf[0] = '\0';
      line_len = 0;

      do {
	if (line_size - line_len < 100)
	  line_buf = realloc(line_buf, line_size += 100);
	fgets(line_buf+line_len, line_size-line_len, fp);
	line_len = strlen(line_buf);
      } while (!feof(fp) && (line_buf[line_len - 1] != '\n'));

      if (!line_len)
	break;
      else if (line_buf[0] == '#') {
	/* we have a program name. parse the buffer and set the program */
	if (all_space(code_buf)) {
	  /* display message and reset the code buffer */
	  sprintf(tmp, "program %s is empty!\n", line_buf+1);
	  (displaystr)(get_buffer(current_buffer), tmp);
	  code_buf[0] = '\0';
	  code_len = 0;
	  continue;
	}

	/* remove the NL at the end. */
	if (line_buf[line_len-1] == '\n')
	  line_buf[line_len-1] = '\0';

	if ((root = my_parse(get_buffer(current_buffer), code_buf)))
	  set_prog(line_buf+1, root, 0, NULL);
	else {
	  sprintf(tmp, "program %s had errors\n", line_buf+1);
	  (displaystr)(get_buffer(current_buffer), tmp);
	}

	/* reset the code buffer */
	code_buf[0] = '\0';
	code_len = 0;
      } else {
	if (code_len+line_len > code_size)
	  code_buf = (char *)realloc(code_buf, code_size += 1000);

	strcpy(code_buf+code_len, line_buf);
	code_len += line_len;
      }
    } while (!feof(fp));

    if (!all_space(code_buf)) {
      /* we have initialization code to be executed immediately */
      if ((root = my_parse(get_buffer(current_buffer), code_buf))) {
	p = NEW(program);
	p->base = root;
	p->ref = 0;
	p->debug = 1;
	p->name = mystrdup("* internal *");
	p->evaled = 1;
	push_task(get_buffer(current_buffer), p, NULL);
      } else {
	(displaystr)(get_buffer(current_buffer),
		     "parse error in stalk file\n");
      }
    }

    free(code_buf);
    free(line_buf);
    fclose(fp);
  }

  found = func_exists(global_funcs, "start");

  if (!found && initlen != 2) /* no start verb & no host/port or pathname/-1 */
    usage(argv[0]);
  if (!found)                 /* no start verb. use host/port from cmd line */
    new_cstream(get_buffer(current_buffer), init[0], atoi(init[1]));
  else {                      /* run start verb */
    v = newlist(initlen);
    for (x=0; x<v->num; x++)
      v->list[x] = newstring(mystrdup(init[x]));
    p = find_program(global_funcs, "start");
    push_task(get_buffer(current_buffer), p, v);
    var_free(v);
  }

  while(!done) {
    tasks_waiting = check_tasks();

    if (tasks_waiting)
      check_streams(MAX_TIMEOUT);
    else
      check_streams(MAX_POLL);
  }

  (end_screen)();

  return 1;
}

/* wrapper for quit */
void w_quit(buf)
     buffer *buf;
{
  quit(0);
}

void quit(sig)
     int sig;
{
  if (sig == SIGSEGV)
    die("Segmentation Fault\n");

  if (sig == SIGILL)
    die("Illegal Instruction\n");

#ifdef SIGBUS
  if (sig == SIGBUS)
    die("Bus Error\n");
#endif /* SIGBUS */

  done = TRUE;
}

/* wrapper for handle_interrupt */
void w_interrupt(buf)
     buffer *buf;
{
  handle_interrupt(0);
}

void handle_interrupt(sig)
     int sig;
{
  (end_screen)();
  kill(getpid(), SIGSTOP);

  (setup_screen)();

  signal(SIGTSTP, handle_interrupt);
  refresh(NULL);
}

void parse_args(argc, argv)
     int argc;
     char **argv;
{
  int x;
  debug_dump = FALSE;

  for (x=1; x<argc; x++) {
    if (!strcmp("-d", argv[x]))
      screen_mode = TERM_DUMB;
    else if (!strcmp("-t", argv[x]))
      screen_mode = TERM_TERMCAP;
#ifdef HAVE_X11
    else if (!strcmp("-x", argv[x]))
      screen_mode = TERM_X;
#endif /* HAVE_X11 */
    else if (!strcmp("-boff", argv[x]))
      set_bell(0);
    else if (!strcmp("-bon", argv[x]))
      set_bell(1);
    else if (!strcmp("-f", argv[x])) {
      if (x == argc)
        usage(argv[0]);
      stalkfile = argv[++x];
    } else if (!strcmp("-nf", argv[x])) {
      stalkfile = "/dev/null";
    } else if (!strcmp("-s", argv[x])) {
      stick_around = TRUE;
    } else if (!strcmp("-g", argv[x])) {
      debug_dump = TRUE;
    } else
      strcpy(init[initlen++], argv[x]);
  }
}

void usage(progname)
     char *progname;
{
  fprintf(stderr, "Usage: %s [-d|-t|-x] [-bon|-boff] [-p|-np] [-s] [-f <stalkfile>|-nf] [-g] args\n", progname);
  fprintf(stderr, "-d  dumb terminal mode (default)\n");
  fprintf(stderr, "-t  full screen (termcap) mode\n");
  fprintf(stderr, "-x  X mode (not implemented yet)\n");
  fprintf(stderr, "-b  turns the bell on/off\n");
  fprintf(stderr, "    upon connection\n");
  fprintf(stderr, "-f  specifies a file to use instead of .stalk.\n");
  fprintf(stderr, "-nf specifies to not use any stalkfile.\n");
  fprintf(stderr, "-s  mean to stick around after the connection is closed.\n");
  fprintf(stderr, "-g  mean to generate debug dumps (if compiled with debugging)\n");
  fprintf(stderr, "%s first looks for a stalkfile. If a -f argument is given, that is\n", progname);
  fprintf(stderr, "used. If not, %s looks for the environmental variable STALK.\n", progname);
  fprintf(stderr, "If that isn't found, %s looks for a file ~/.stalk.\n", progname);
  fprintf(stderr, "If a stalkfile is found, it is parsed as a series of commands and program\n");
  fprintf(stderr, "declarations. If it defines a program called 'start', then start is run,\n");
  fprintf(stderr, "and is given 'args' (from the command line) as arguments\n");
  fprintf(stderr, "If 'start' is not defined, then 'args' must consist of a hostname followed by\n");
  fprintf(stderr, "a port number or a pathname followed by -1 (to use UNIX domain sockets).\n");
  (end_screen)();
  exit (0);
}

#ifdef DEBUG_MEM
void setup_memcheck()
{
  union dbmalloptarg v;

  v.str = "memlog";
  dbmallopt(MALLOC_ERRFILE, &v);
}
#endif /* DEBUG_MEM */
