/*
 * stalk/stream.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "stream.h"
#include "stuff.h"
#include "functions.h"
#include "var.h"
#include "buffer.h"
#include "parser.h"
#include "interpret.h"
#include "task.h"

#define MAX_STREAMS 256

static int ms_per_tick = 0;
static struct tms naar;

void init_streams(buf)
     buffer *buf;
{
  stream *s;
  int ticks_per_sec;

  if (!ms_per_tick) {
    if ((ticks_per_sec = sysconf(_SC_CLK_TCK)) == 0)
      die("stream.c: sysconf(_SC_CLK_TCK) error! (line %d)", __LINE__);
    ms_per_tick = 1000/ticks_per_sec;
    times(&naar);
  }

  buf->next_stream = 0;
  buf->streams = NULL;
  s = newstream(NULL, buf);
  s->type = stream_default;
}

stream *newstream(t, buf)
     task *t;
     buffer *buf;
{
  /*
   * creates a new stream owned by the current task
   * it is up to the calling procedure to open the appropriate Socket,
   * FILE *, whatever and set the appropriate flags in the structure.
   * (on_close, on_input, filep, sock, and type)
   */

  stream *s1, *s2;

  s1 = NEW(stream);
  s1->type = stream_none;
  s1->streamnum = buf->next_stream++;

  if (t)
    s1->task_id = t->task_id;
  else
    s1->task_id = ALL_TASKS;

  s1->on_close = NULL;
  s1->on_input = NULL;

  if (t)
    s1->process_input = 0; /* the task can set it itself */
  else
    s1->process_input = 1;

  s1->input_mode = stream_text;
  s1->snarf_output = 0;

  s1->sock = NULL;
  s1->filep = NULL;

  s1->buf = buf;

  s1->timeout = 0;
  s1->timeleft = 0;

  s1->waiting_task = NULL;

  s1->buffer = NULL;
  s1->buf_size = 0;
  s1->num_read = 0;

  s1->next = NULL;
  s2 = buf->streams;
  if (s2) {
    while (s2->next)
      s2 = s2->next;
    s2->next = s1;
    s1->prev = s2;
  } else {
    s1->prev = NULL;
    buf->streams = s1;
  }

  return s1;
}

stream *findstream(t, n)
     task *t;
     int n;
{
  /*
   * returns stream referenced by n if the stream exists and is
   * accessable by the current task.
   * returns NULL otherwise.
   */
  stream *s;

  s = t->buf->streams;
  while (s) {
    if (s->streamnum == n) {
      if ((s->task_id == t->task_id) || (s->task_id == ALL_TASKS))
	return s;
      else
	return NULL;
    }
    s = s->next;
  }

  return NULL;
}

void freestream(s)
     stream *s;
{
  if (!s)
    return;

  switch (s->type) {
  case stream_none:
    break;
  case stream_default:
    break;
  case stream_sock_serv:
  case stream_sock_clnt:
    sock_close(&(s->sock));
    break;
  case stream_file:
  case stream_pipe:
    if (s->filep)
      fclose(s->filep);
    break;
  default:
    return;
    break;
  }

  if (s->next)
    s->next->prev = s->prev;

  if (s->prev)
    s->prev->next = s->next;

  free(s);
}

void free_task_streams(buf, task_id)
     buffer *buf;
     int task_id;
{
  stream *s;

  s = buf->streams;
  while (s->next) {
    if (s->next->task_id == task_id)
      stream_close(s->next);
    else
      s = s->next;
  }
}

void stream_close(s)
     stream *s;
{
  var *v;
  program *p;

  if (s->waiting_task) {
    if (s->num_read)
      process_input(s);
    else {
      in_push(s->waiting_task, newnum(-1));
      s->waiting_task->type = TA_READY;
    }
  }

  if (s->on_close && func_exists(global_funcs, s->on_close)) {
    v = newlist(1);
    v->list[0] = newnum(s->streamnum);
    p = find_program(global_funcs, s->on_close);
    push_task(s->buf, p, v);
    var_free(v);
  }

  freestream(s);
}

void check_streams(maxtimeout)
     int maxtimeout;
{
  int x, num_streams, fds[MAX_STREAMS], timeout;
  char ready[MAX_STREAMS];
  stream *streamp[MAX_STREAMS], *temp;

  timeout = maxtimeout;

  num_streams = 0;
  for (x=0; x<num_buffers; x++)
    if (buffers[x]) {
      temp = buffers[x]->streams;
      while (temp->next) {
	if ((temp->next->process_input || temp->next->waiting_task)
	    && (temp->type != stream_file))
	  streamp[num_streams++] = temp->next;
	temp = temp->next;
      }
    }

  for (x=0; x<num_streams; x++) {
    switch (streamp[x]->type) {
    case stream_sock_serv:
    case stream_sock_clnt:
      fds[x] = sock_fd(streamp[x]->sock);
      break;
    case stream_file:
    case stream_pipe:
      fds[x] = fileno(streamp[x]->filep);
      break;
    default:
      fds[x] = -1;
      break;
    }
  }


  streamp[num_streams] = buffers[current_buffer]->streams;
  fds[num_streams++] = FD_IN;

  sock_select(fds, num_streams, ready, timeout);

  for (x=0; x<num_streams; x++) {
    if (ready[x])
      stream_read(streamp[x]);
    else {
      if (streamp[x]->timeout) {
	debit_stream_time(streamp[x]);
	if (streamp[x]->timeleft <= 0)
	  process_input(streamp[x]);
      }
    }
  }
}

void process_input(s)
     stream *s;
{
  var *v;
  program *p;
  statement *stmt;

  s->buffer[s->num_read] = '\0';

  if (s->waiting_task) {
    /*
     * a task is doing a read() on this stream.
     * send the line if something was read. send 0 if nothing was read
     */
    if (s->num_read) {
      in_push(s->waiting_task, newstring(s->buffer));
      s->buffer = NULL;
      s->buf_size = 0;
      s->waiting_task->type = TA_READY;
    } else {
      in_push(s->waiting_task, newnum(0));
      s->waiting_task->type = TA_READY;
    }
  } else if (s->num_read) {
    if (s->on_input && func_exists(global_funcs, s->on_input)) {
      /* start up a new task */
      v = newlist(1);
      v->list[0] = newstring(s->buffer);
      s->buffer = NULL;
      s->buf_size = 0;
      p = find_program(global_funcs, s->on_input);
      push_task(s->buf, p, v);
      var_free(v);
    } else {
      if (s->type == stream_default) {
	/*
	 * line typed from the keyboard.
	 * either interpret it (if it has one leading '/' and is 2 or more
	 * characters) or send it to all snarfing sockets
	 */

	if (s->buffer[0] == '/') {
	  /*
	   * if it starts with '//', the first '/' and send it to all
	   * output streams. if it is just '/', send it anyway.
	   */

	  if (s->num_read == 1)
	    send_to_streams(s->buf, s->task_id, s->buffer);
	  else if (s->buffer[1] == '/')
	    send_to_streams(s->buf, s->task_id, s->buffer+1);
	  else {
	    /* eval it */
	    stmt = my_parse(s->buf, s->buffer);
	    if (stmt) {
	      p = NEW(program);
	      p->base = stmt;
	      p->ref = 0;
	      p->evaled = 1;
	      p->debug = 1;
	      p->name = mystrdup("* keyboard program *");
	      push_task(s->buf, p, NULL);
	    }
	  }
	} else
	  send_to_streams(s->buf, s->task_id, s->buffer);
      } else
	(displaystr)(s->buf, s->buffer);
    }
  }

  s->num_read = 0;
  s->waiting_task = NULL;
}

void debit_stream_time(s)
     stream *s;
{
  int ticks_passed, time_passed;
  clock_t check_time;

  if ((s->waiting_task || s->num_read) && s->timeout) {
    ticks_passed = (check_time = times(&naar)) - s->lastcheck;
    s->lastcheck = check_time;
    time_passed = ticks_passed * ms_per_tick;
    s->timeleft -= time_passed;
  }
}

void stream_read(s)
     stream *s;
{
  int timeout; /* timeout in milliseconds */
  int read_here; /* counts the number of characters read in this invocation
		  * of stream_read */
  int c;
  Socket sock;
  stream *s2;
  var *v;
  program *p;

  if (s->type == stream_sock_serv) {
    if (s->waiting_task) {
      s2 = newstream(s->waiting_task, s->buf);
      if (!s2) {
	in_push(s->waiting_task, newnum(-1));
	s->waiting_task->type = TA_READY;
	s->waiting_task = NULL;
	return;
      }
      if (sock_accept(s->sock, &sock, 10) == -1) {
	stream_close(s);
	freestream(s2);
	in_push(s->waiting_task, newnum(-1));
	s->waiting_task->type = TA_READY;
	s->waiting_task = NULL;
	return;
      }
      s2->type = stream_sock_serv;
      s2->sock = sock;
      in_push(s->waiting_task, newnum(s2->streamnum));
      s->waiting_task->type = TA_READY;
      s->waiting_task = NULL;
      return;
    }

    if (!(s2 = newstream(s->waiting_task, s->buf)))
      return;
    if (sock_accept(s->sock, &sock, 10) == -1) {
      stream_close(s);
      freestream(s2);
      return;
    }
    s2->type = stream_sock_serv;
    s2->sock = sock;
    if (s->on_input && func_exists(global_funcs, s->on_input)) {
      /* start up a new task */
      v = newlist(2);
      v->list[0] = newnum(s->streamnum);
      v->list[1] = newnum(s2->streamnum);
      p = find_program(global_funcs, s->on_input);
      push_task(s->buf, p, v);
      var_free(v);
    }
    return;
  }

  if (!s->buf_size) {
    s->buffer = (char *)malloc(s->buf_size = 100);
    s->num_read = 0;
  }

  if ((s->type == stream_default) && !s->waiting_task && s->process_input) {
    /*
     * we're supposed to process input and perform the default action on
     * the keyboard. In this case the timeout doesn't matter. just
     * build up a string with all the characters we can get and pass it
     * to the appropriate terminal driver.
     */

    while ((c = stream_readchar(s, 1)) != -1) {
      if (s->num_read+2 > s->buf_size)
	s->buffer = (char *)realloc(s->buffer, (s->buf_size += 25));
      s->buffer[s->num_read++] = c;
    }

    s->buffer[s->num_read] = '\0';
    check_functions(get_buffer(current_buffer), s->buffer);
    s->num_read = strlen(s->buffer);
    return;
  }

  read_here = 0;
  timeout = READ_TIMEOUT;

  while (1) {
    c = stream_readchar(s, timeout);
    if (c != EOF) {
      read_here++;
      s->timeleft = s->timeout;
      timeout = MIN(s->timeleft, READ_TIMEOUT);
      if (timeout == 0)
	timeout = READ_TIMEOUT;

      /* append the character to the buffer */
      if ((s->input_mode == stream_text) && (c == '\r'))
	/* discard the character */;
      else {
	if (s->num_read + 2 >= s->buf_size)
	  s->buffer = (char *)realloc(s->buffer, (s->buf_size += 25));
	s->buffer[s->num_read++] = c;
      }

      if ((s->input_mode == stream_text) && (c == '\n')) {
	process_input(s);
	return;
      } else if ((s->input_mode == stream_raw) && (s->num_read == s->nchars)) {
	process_input(s);
	return;
      }
    } else {
      if (!read_here) {
	if (s->num_read)
	  process_input(s);
	stream_close(s);
	return;
      }
      s->timeleft -= timeout;
      if (s->timeout && (s->timeleft <= 0))
	process_input(s);

      /* nothing was read, so return */
      s->lastcheck = times(&naar);
      return;
    }
  }
}

int stream_readchar(s, timeout)
     stream *s;
     int timeout;
{
  struct timeval Timeout;
  int fd;
  fd_set readfds;
  char c;

  switch (s->type) {
  case stream_default:
    fd = FD_IN;
    break;
  case stream_sock_serv:
  case stream_sock_clnt:
    fd = sock_fd(s->sock);
    break;
  case stream_file:
    return fgetc(s->filep);
    break;
  case stream_pipe:
    fd = fileno(s->filep);
    break;
  default:
    return EOF;
    break;
  }

  if (timeout) {
    Timeout.tv_usec = (timeout % 1000) * 1000;
    Timeout.tv_sec = timeout / 1000;
  }

  FD_ZERO(&readfds);
  FD_SET(fd, &readfds);
  if (select(fd+1, &readfds, NULL, NULL, timeout ? &Timeout : NULL) < 0)
    return EOF;

  if (!FD_ISSET(fd, &readfds))
    return EOF;

  if (read(fd, &c, 1) != 1)
    return EOF;
  else
    return c;
}

/*
 * will send the string to all snarfing streams on buf.
 * if the string came from a stream, send the stream as s also,
 * else make s NULL (if it came from the keyboard or something).
 */
void send_to_streams(buf, task_id, string)
     buffer *buf;
     int task_id;
     char *string;
{
  stream *temp;

  for (temp = buf->streams; temp; temp = temp->next) {
    if ((temp->task_id != ALL_TASKS) && (temp->task_id != task_id))
      continue;

    if (!temp->snarf_output)
      continue;

    switch (temp->type) {
    case stream_default:
      /* 
       * we don't want to send keyboard input to the display a
       * second time
       */
      break;
    case stream_sock_clnt:
      sock_printf(temp->sock, "%s", string);
      break;
    case stream_file:
    case stream_pipe:
      fprintf(temp->filep, "%s", string);
      break;
    default:
      /* D'oh! */
      break;
    }
  }
}

int new_cstream(buf, name, port)
     buffer *buf;
     char *name;
     int port;
{
  stream *s;

  if (!(s = newstream(NULL, buf)))
    return -1;

  s->type = stream_sock_clnt;

  if (sock_call_server(&(s->sock), name, port) == -1) {
    freestream(s);
    return -1;
  }
  s->snarf_output = 1;
  return 0;
}

/*
 * builtin functions from the PL that act on streams
 */

void bi_popen(t, args)
     task *t;
     var *args;
{
  stream *s;
  static var_type types[] = { V_STRING, V_LIST };
  Error e;
  char **argv;
  int argc, i;

  if ((e = inv_args(args, 1, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->num == 1 || !args->list[1]->num)
    argc = 0;
  else {
    argc = args->list[1]->num;
    for (i=0; i<argc; i++)
      if (args->list[1]->list[i]->type != V_STRING) {
	in_push(t, make_error(t, E_TYPE));
	return;
      }
    argv = (char **)malloc((argc+1)*sizeof(char *));
    for (i=0; i<argc; i++)
      argv[i] = args->list[1]->list[i]->string;
    argv[argc] = NULL;
  }

  s = newstream(t, t->buf);
  s->type = stream_pipe;

  /***/ /* fork and exec stream */

  in_push(t, newnum(s->streamnum));
}

void bi_fopen(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_STRING, V_STRING };
  Error e;
  stream *s;
  char *filename;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  s = newstream(t, t->buf);
  s->type = stream_file;

  filename = expand_filename(args->list[0]->string);
  if ((s->filep = fopen(filename, args->list[1]->string)) == NULL) {
    freestream(s);
    in_push(t, newnum(-1));
  } else
    in_push(t, newnum(s->streamnum));

  if (filename)
    free(filename);
}

void bi_copen(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_STRING, V_NUM };
  Error e;
  stream *s;
  int port;

  if ((e = inv_args(args, 1, 2, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  s = newstream(t, t->buf);
  s->type = stream_sock_clnt;

  if (args->num == 2)
    port = args->list[1]->num;
  else
    port = SOCK_UNIXDOM;

  if (sock_call_server(&(s->sock), args->list[0]->string, port) == -1) {
    freestream(s);
    in_push(t, newnum(-1));
  } else
    in_push(t, newnum(s->streamnum));
}

void bi_sopen(t, args)
     task *t;
     var *args;
{
  stream *s;
  int result;

  if (args->num != 1) {
    in_push(t, make_error(t, E_ARGS));
    return;
  } else if ((args->list[0]->type != V_NUM) &&
	     (args->list[0]->type != V_STRING)) {
    in_push(t, make_error(t, E_TYPE));
    return;
  }

  s = newstream(t, t->buf);
  s->type = stream_sock_clnt;

  if (args->list[0]->type == V_STRING) {
    result = sock_make_server(&(s->sock), args->list[0]->string, SOCK_UNIXDOM);
  } else {
    if (args->list[0]->num > 0)
      result = sock_make_server(&(s->sock), NULL, args->list[0]->num);
    else
      result = -1;
  }

  if (result == -1) {
    freestream(s);
    in_push(t, newnum(-1));
  } else
    in_push(t, newnum(s->streamnum));
}

void bi_accept(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;
  stream *s;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    if (s->waiting_task)
      in_push(t, newnum(-1));
    else {
      s->waiting_task = t;
      t->type = TA_READING;
      /*
       * don't push anything onto the task's stack. a value will be pushed by
       * the read loop
       */
    }
  }
}

void bi_close(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;
  stream *s;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    freestream(s);
    in_push(t, newnum(0));
  }
}

void bi_read(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;
  stream *s;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    if (s->waiting_task)
      in_push(t, newnum(-1));
    else {
      s->waiting_task = t;
      t->type = TA_READING;
      /*
       * don't push anything onto the task's stack. a value will be pushed by
       * the read loop
       */
    }
  }
}

void bi_write(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM, V_STRING };
  Error e;
  stream *s;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    switch (s->type) {
    case stream_none:
      in_push(t, newnum(-1));
      break;
    case stream_default:
      (displaystr)(t->buf, args->list[1]->string);
      in_push(t, newnum(0));
      break;
    case stream_sock_serv:
      in_push(t, newnum(-1));
      break;
    case stream_sock_clnt:
      sock_printf(s->sock, "%s", args->list[1]->string);
      in_push(t, newnum(0));
      break;
    case stream_file:
      fprintf(s->filep,  "%s", args->list[1]->string);
      fflush(s->filep);
      in_push(t, newnum(0));
      break;
    case stream_pipe:
      fprintf(s->filep,  "%s", args->list[1]->string);
      in_push(t, newnum(0));
      break;
    default:
      in_push(t, newnum(-1));
      break;
    }
  }
}

void bi_set_timeout(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM, V_NUM };
  Error e;
  stream *s;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    s->timeout = args->list[1]->num;
    in_push(t, newnum(0));
  }
}

void bi_capture_output(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM, V_NUM };
  Error e;
  stream *s;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    s->snarf_output = args->list[1]->num;
    in_push(t, newnum(0));
  }
}

void bi_export(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;
  stream *s;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    s->task_id = ALL_TASKS;
    in_push(t, newnum(0));
  }
}

void bi_on_input(t, args)
     task *t;
     var *args;
{
  stream *s;

  if (args->num != 2) {
    in_push(t, make_error(t, E_ARGS));
    return;
  } else if ((args->list[0]->type != V_NUM) ||
	     ((args->list[1]->type != V_NUM) &&
	      (args->list[1]->type != V_STRING))) {
    in_push(t, make_error(t, E_TYPE));
    return;
  }

  if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    if (s->on_input) {
      free(s->on_input);
      s->on_input = NULL;
    }

    if (args->list[1]->type == V_NUM)
      s->process_input = args->list[1]->num;
    else {
      s->process_input = 1;
      if (strlen(args->list[1]->string))
	s->on_input = mystrdup(args->list[1]->string);
      else
	s->on_input = NULL;
    }

    in_push(t, newnum(0));
  }
}

void bi_on_close(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM, V_STRING };
  Error e;
  stream *s;

  if ((e = inv_args(args, 2, 2, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else if (!(s = findstream(t, args->list[0]->num)))
    in_push(t, newnum(-1));
  else {
    if (s->on_close)
      free(s->on_close);

    if (strlen(args->list[1]->string))
      s->on_close = mystrdup(args->list[1]->string);
    else
      s->on_close = NULL;

    in_push(t, newnum(0));
  }
}

void bi_broadcast(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_STRING };
  Error e;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else {
    send_to_streams(t->buf, t->task_id, args->list[0]->string);

    in_push(t, newnum(0));
  }
}

void bi_set_mode(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}

void bi_streams(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}

void bi_stream_ready(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}
