/*
 * stalk/structures.h
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef ___Struc__
#define ___Struc__

#include "shockit.h"

typedef void (*key_func) (/* struct buffer *buf */);

/*
 * stream_default is the keyboard (for input)/current buffer (for output)
 * it has a descriptor of 0
 */
typedef enum {
  stream_none,
  stream_default, stream_sock_serv, stream_sock_clnt, stream_file, stream_pipe
} streamtype;

typedef enum {
  stream_raw, stream_text
} stream_mode;

typedef struct stream {
  streamtype type;

  /* unique number assigned to each connection.
   * this is used by the stalkcode program to access the stream */
  int streamnum;

  /* this is that task associated with the stream. if it is -1, then the
   * task has exported it to the buffer. if it is owned by a task, it
   * will be closed as soon as the task's thread ends. */
  int task_id;

  FILE *filep;
  Socket sock;

  char *on_close; /* function to call upon detecting a close on the stream */

  int process_input; /* should input be processed or should we let the
		      * user request it explicitly?
		      * default = process it */
  char *on_input; /* function to call upon detecting input from the stream
		   * default = send it to the screen (no func specified) */
  stream_mode input_mode; /* text or raw?
			   * default = text */
  int snarf_output; /* get output from unhooked streams?
		     * default = yes, snarf output */
  int nchars; /* number of characters to be read if in raw mode */

  /* the buffer that the stream was opened on */
  struct buffer *buf;

  /* timeout in milliseconds (if >0), block if == 0 */
  int timeout;

  /* this must be NULL if no task is doing a read() on the stream */
  struct task *waiting_task;

  /* the buffer of characters that have been read in */
  char *buffer;
  int buf_size;
  int num_read;

  /* time remaining before the read buffer times out and must be processed */
  int timeleft;
  clock_t lastcheck;

  struct stream *next, *prev;
} stream;

typedef enum {
  binding_none,
  binding_prefix,
  binding_print,
  binding_builtin,
  binding_stalkcode
} binding_type;

typedef struct binding {
  binding_type type;  /* is the binding invalid, a prefix,
		       * print the character,
		       * or a built-in or interpreted function? */
  key_func function;  /* function pointer if builtin */
  char *name;         /* name of function if not builtin */
  struct binding **prefix; /* prefix table for the next character */
} binding;

typedef struct line {
  struct line *next, *prev;
  char *string;
  int eol;
} line;

typedef struct {
  line *lines;   /* two-way-list head */
  line *cursor;  /* used to traverse the two-way-list of lines */
  int size;
  int pos;
} line_queue;

typedef struct {
  char *function_name;
  key_func function;
} function_binding;

typedef struct {
  char line[2500];
  int len, pos;
} keyboard_state;

typedef struct buffer {
  int buf_id; /* each buffer has a unique id */
  struct var_env *global;

  line_queue *lines, *history;
  int history_pos;

  binding **local_bindings; /* table of local bindings.
			   * these are used before global bindings */

  function_binding *function_bindings;

  char *name; /* buffer name */

  int scrollmode, connected; /* status flags */

  int startline;     /* the line of stdscr that the window starts on */
  int height;        /* size of the window - all windows have width = COLS */

  /*
   * row - screen line that the current input line starts on
   * 0 means it starts on the top line of the screen. if this is < startline,
   * then the line being typed is larger than the current buffer's screensize
   */
  int row;

  keyboard_state keyboard;

  int next_stream;
  stream *streams;
} buffer;



/* program types */

/* MN_OR1 and MN_AND1 are 'pre' ors and ands for the short-circuit evaluation.
 * they don't get seen in the unparsing. they are needed due to the fact that
 * programs are stored in postfix form. */

typedef enum {
  MN_NONE, MN_NOOP, MN_JUMP,
  MN_IF, MN_ELSEIF, MN_ELSE, MN_ENDIF,
  MN_WHILE, MN_ENDWHILE,
  MN_EXIT, MN_RETURN, MN_EMPTY_RETURN,
  MN_FOR_LIST, MN_FOR_RANGE, MN_ENDFOR,
  MN_FORK, MN_ENDFORK,
  MN_ID, MN_LOCAL_VAR, MN_GLOBAL_VAR,
  MN_VALUE, MN_ENDEXPR,
  MN_ELEMENT, MN_ELEMENT_RANGE,
  MN_LIST_ELEMENT, MN_LIST_COMBINE, MN_LIST_SPLICE, MN_LIST_EMPTY, MN_LIST,
  MN_LOCAL_ASSIGN, MN_GLOBAL_ASSIGN, MN_END_ASSIGN,
  MN_FUNCTION, MN_INDIRECT_FUNCTION, MN_CONDITION,
  MN_EQ, MN_NE, MN_GE, MN_GT, MN_LE, MN_LT,
  MN_ADD, MN_SUB, MN_MUL, MN_DIV, MN_MOD, MN_NEG,
  MN_OR, MN_AND, MN_NOT, MN_OR1, MN_AND1, MN_IN,
  MN_LINE,  /* not actually executed, this statement
	     * actually holds the unparsed text of the
	     * currently executing line */
  MN_DEBUG, MN_END_DEBUG, MN_CALL, MN_IGNORE, MN_FATAL
} mnemonic;

/*
 * if you add any errors, please change NUM_ERRORS and add the error message
 * to init_vars in var.c
 */
#define NUM_ERRORS 11

typedef enum {
  E_NONE,
  E_VARNF, E_TYPE, E_DIV, E_VERBNF, E_MAXREC, E_RANGE, E_ARGS, E_INVARG,
  E_EXIT, E_TICKS
} Error;

typedef enum {
  V_NONE, V_STRING, V_LIST, V_NUM, V_ERR, V_LABEL, V_TOKEN, V_INDEX
} var_type;

typedef struct var {
  var_type type;
  int ref;
  char *string;
  struct var **list;
  int num;
  Error err;
} var;

typedef struct statement {
  mnemonic type;
  struct statement *next, *jump;
  var *value;
} statement;

typedef struct {
  statement *base;
  int ref;     /* reference count for the number of times this program
		* appears in the task list. Used to check when evaled
		* programs should be removed and if it is safe to
		* remove a verb that is being redefined. */
  int debug;
  char *name;
  int evaled;
} program;

typedef struct {
  char *string;
  int value;
  Error err;
} tokenentry;

typedef struct var_env {
  char *name;
  var *value;
  struct var_env *left, *right;
} var_env;

typedef enum {
  TA_NONE, TA_READY, TA_WAITING, TA_SUSPENDED, TA_FORKED, TA_SENT_EVENT,
  TA_READING, TA_DYING, TA_RETURNING
} task_type;

typedef struct var_stack {
  var *val;
  struct var_stack *next;
} var_stack;

typedef struct task {
  task_type type;
  int task_id;
  int counter;
  time_t wakeup;
  int eval;  /* whether a sent buffer event should be evaled or not */
  buffer *buf;
  int level; /* the level of recursion */

  program *program;
  statement *current;
  var_env *local;

  char *buffer_event;

  var *line_info; /* reference to the line info from the MN_LINE statement */
  var_stack *stack;

  int ignore_errors; /* true if in the error part of an on_error statement */
  int debug_task; /*
		   * true if this task is a debug task created because of
		   * an error condition. this means the task does not
		   * get to return a value
		   */

  int debug[NUM_ERRORS];
  char *debug_func[NUM_ERRORS];
  int error_pending;
  Error error;

#ifdef EVIL_DEBUGGING_MODE
  FILE *dumpfile;
#endif

  struct task *next, *prev, *parent, *child;
} task;

typedef void (*bi_func) P((task *t, var *args));

typedef struct prog_env {
  char *name;
  int builtin;
  bi_func func;
  program *program;
  struct prog_env *left, *right;
} prog_env;

typedef struct {
  int exit_status;
  int return_status;
  var *return_value;
} returnval;



/* parser/unparser structures */

typedef struct string_list {
  char *string;
  struct string_list *next;
} string_list;

typedef string_list unparse_stack;

typedef struct {
  statement *s;
  int lhs;

  /*
   * variables that are accessed in this expression.
   * needs to be known for the debugger
   */
  string_list *lvars;
  string_list *gvars;
  char *upexpr; /* unparsed form of the expression. */
} exprstruct;

#endif /* ___Struc__ */
