/*
 * stalk/task.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "task.h"
#include "var.h"
#include "stuff.h"
#include "interpret.h"
#include "stream.h"
#include "functions.h"
#include "parser.h"
#include "debug.h"

static task *tasks = NULL;
static int next_task_id = 1;
int starting_ticks = MAX_TICK;

/*
 * task_free will free the rest of the interpreter stack, all the
 * current task's allocated storage, and destroy the task, bumping the
 * next one up in line. v is the return value of the current task.
 */
void task_free(t)
     task *t;
{
  var_stack *stack;

  if (t->type == TA_SENT_EVENT) {
    free(t->buffer_event);
  } else {
    t->program->ref--;
    if (t->program->evaled && !t->program->ref)
      free_program(t->program);
    free_env(t->local);
    while(t->stack) {
      var_free(t->stack->val);
      stack = t->stack;
      t->stack = t->stack->next;
      free(stack);
    }
  }

  if (t->parent) {
    /* remove the task and put it's parent in it's place */
    t->parent->next = t->next;
    t->parent->prev = t->prev;
    t->prev->next = t->parent;
    if (t->next)
      t->next->prev = t->parent;

    /*
     * if the task is returning, wake the parent.
     * if it's doing something like dying, have the parent inherit
     * that state.
     */
    if (t->type == TA_RETURNING && t->parent->type == TA_WAITING)
      t->parent->type = TA_READY;
    else
      t->parent->type = t->type;

    /* update the parent's tick counter */
    if ((t->counter >= 0) && (t->parent->counter >= t->counter))
      t->parent->counter = t->counter;
  } else {
    /* remove the task and chain the rest together */
    if (t->next)
      t->next->prev = t->prev;
    t->prev->next = t->next;
    free_task_streams(t->buf, t->task_id);
  }

#ifdef EVIL_DEBUGGING_MODE
  /* if we are debugging, close the dumpfile */
  if (t->dumpfile)
    fclose(t->dumpfile);
#endif /* EVIL_DEBUGGING_MODE */

  free(t);
}

void init_tasks()
{
  if (tasks)
    return;

  tasks = NEW(task); /* dummy head for the task table */
  tasks->parent = tasks->child = NULL;
  tasks->prev = tasks->next = NULL;
  tasks->task_id = -1;
  tasks->type = TA_NONE;
}

void exec_builtin(penv, name, t, args)
     prog_env *penv;
     char *name;
     task *t;
     var *args;
{
  int c;

  if (!penv)
    return;

  if (!(c = strcmp(penv->name, name)) && !penv->builtin)
    return;
  else if (!c)
    (penv->func)(t, args);
  else if (c < 0)
    exec_builtin(penv->right, name, t, args);
  else
    exec_builtin(penv->left, name, t, args);
}

/*
 * exec_task requires the program * for a task.
 * if you only know the name, use find_program to get the statement *
 */
void exec_task(t, p, args, debug)
     task *t;
     program *p;
     var *args;
     int debug;
{
  task *ntask;

  ntask = NEW(task);
  ntask->level = t->level+1;
  ntask->next = t->next;
  ntask->prev = t->prev;
  ntask->prev->next = ntask;
  if (ntask->next)
    ntask->next->prev = ntask;
  ntask->parent = t;
  t->child = ntask;
  ntask->child = NULL;

  ntask->type = TA_READY;
  ntask->task_id = t->task_id;
  ntask->program = p;
  ntask->current = p->base;
  ntask->local = NEW(var_env);
  ntask->local->name = mystrdup("args");
  ntask->local->value = var_ref(args);
  ntask->local->left = ntask->local->right = NULL;
  ntask->buf = t->buf;
  ntask->debug_task = debug;
  if (t->counter == -1)
    ntask->counter = starting_ticks;
  else
    ntask->counter = t->counter;
  ntask->stack = NULL;
  ntask->wakeup = 0;
  ntask->buffer_event = NULL;
  ntask->eval = 0;

#ifdef EVIL_DEBUGGING_MODE
  ntask->dumpfile = NULL;
#endif /* EVIL_DEBUGGING_MODE */

  init_debug(ntask);

  t->type = TA_WAITING;

  p->ref++;
}

/*
 * creates a blank task at the end of the task table
 * and returns it. anything you put here to initialize things should
 * also be put into exec_task
 */
task *new_task(buf)
     buffer *buf;
{
  task *ntask;

  ntask = tasks;
  while (ntask->next)
    ntask = ntask->next;
  ntask->next = NEW(task);
  ntask->next->prev = ntask;
  ntask = ntask->next;
  ntask->parent = NULL;
  ntask->child = NULL;
  ntask->next = NULL;
  ntask->task_id = next_task_id++;
  ntask->stack = NULL;
  ntask->buf = buf;
  ntask->counter = starting_ticks;
  ntask->wakeup = 0;
  ntask->buffer_event = NULL;
  ntask->eval = 0;

#ifdef EVIL_DEBUGGING_MODE
  ntask->dumpfile = NULL;
#endif /* EVIL_DEBUGGING_MODE */

  init_debug(ntask);

  return ntask;
}

/*
 * push_task puts a task at the end of the waiting queue
 * push_task requires the program * for a task.
 * if you only know the name, use find_program to get the program *
 */
void push_task(buf, p, args)
     buffer *buf;
     program *p;
     var *args;
{
  task *ntask;

  ntask = new_task(buf);
  ntask->level = 1;
  ntask->type = TA_READY;
  ntask->program = p;
  ntask->current = p->base;
  ntask->local = NEW(var_env);
  ntask->local->name = mystrdup("args");
  ntask->local->value = var_ref(args);
  ntask->local->left = ntask->local->right = NULL;

  p->ref++;
}

void send_buffer_event(buf, event, eval)
     buffer *buf;
     char *event;
     int eval;
{
  task *ntask;

  ntask = new_task(buf);
  ntask->level = 1;
  ntask->type = TA_SENT_EVENT;
  ntask->program = NULL;
  ntask->current = NULL;
  ntask->eval = eval;
  ntask->local = NULL;
  ntask->buffer_event = mystrdup(event);
}

/*
 * fork_task puts a task at the end of the waiting queue
 */
void fork_task(buf, p, s, delay)
     buffer *buf;
     program *p;
     statement *s;
     int delay; /* seconds */
{
  task *ntask;

  ntask = new_task(buf);
  ntask->level = 1;
  ntask->type = TA_FORKED;
  ntask->program = p;
  ntask->current = s;
  ntask->local = NEW(var_env);
  ntask->local->name = mystrdup("args");
  ntask->local->value = NULL;
  ntask->local->left = ntask->local->right = NULL;
  ntask->wakeup = time(0) + delay;

  p->ref++;
}

/*
 * check_tasks() will give a timeslice to each waiting task, then return.
 * check_tasks() will return 1 if there are ready tasks, 0 if no tasks are
 * currently ready for another timeslice (all suspended, reading, etc)
 */
int check_tasks()
{
  task *t, *t2, *t3;
  var *v;
  program *p;
  int x;
  statement *temp;

  t = tasks;
  while (t) {
    switch (t->type) {
    case TA_READY:
      /*
       * move task to the head if it wasn't there already
       * (as in the case of a pushed task or the head task suspended)
       */
      if (t != tasks->next)
	wake_task(t);
      else
	ntrp_task(t);
      t = tasks->next;
      break;
    case TA_SUSPENDED:
    case TA_FORKED:
      if (time(0) >= t->wakeup) {
	wake_task(t);
	t = tasks;
      } else
	t = t->next;
      break;
    case TA_SENT_EVENT:
      if (t->eval) {
	if (func_exists(global_funcs, "buffer")) {
	  v = newlist(1);
	  v->list[0] = newstring(t->buffer_event);
	  p = find_program(global_funcs, "buffer");
	  push_task(t->buf, p, v);
	  var_free(v);
	} else {
	  x = strlen(t->buffer_event);
	  t->buffer_event[x++] = '\n';
	  t->buffer_event[x] = '\0';
	  temp = my_parse(t->buf, t->buffer_event);
	  if (temp) {
	    p = NEW(program);
	    p->debug = 1;
	    p->name = mystrdup("* sent event *");
	    p->base = temp;
	    p->ref = 0;
	    p->evaled = 1;
	    push_task(t->buf, p, NULL);
	  }
	}
      } else
	(displaystr)(t->buf, t->buffer_event);
      task_free(t);
      t = tasks;
      break;
    case TA_DYING:
      t3 = t;
      t = t->next;

      /* destroy the task */
      t2 = t3->parent;
      task_free(t3);
      t3 = t2;
      while (t3) {
	t2 = t3->parent;
	task_free(t3);
	t3 = t2;
      }
      break;
    case TA_RETURNING:
      if (t->parent && !t->debug_task) {
	v = in_pop(t);
	in_push(t->parent, v);
	if (v && (v->type == V_ERR))
	  raise_error(t, v->err);
      }

      if (t->parent) {
	t = t->parent;
	task_free(t->child);
      } else if (t->next) {
	t = t->next;
	task_free(t->prev);
      } else {
	task_free(t);
	t = NULL;
      }
      break;
    default:
      t = t->next;
      break;
    }
  }
  return 0;
}

/*
 * move this task and all waiting tasks to the front of the queue
 * so that another suspended task doesn't wake up if this one
 * halts halfway through by execing a function or something
 */
void wake_task(t)
     task *t;
{
  task *temp;

  t->type = TA_READY;
  if (tasks->next == t)
    return;

  temp = tasks->next;
  temp->prev = t;
  tasks->next = t;
  t->prev->next = t->next;
  if (t->next)
    t->next->prev = t->prev;
  t->prev = tasks;
  t->next = temp;
}

/*
 * builtin functions from the PL that act on tasks
 */

void bi_set_ticks(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else {
    t->counter = args->list[0]->num;
    in_push(t, NULL);
  }
}

void bi_default_ticks(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;

  if ((e = inv_args(args, 0, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else {
    in_push(t, newnum(starting_ticks));
    if (args->num)
      starting_ticks = args->list[0]->num;
  }
}

void bi_ticks_left(t, args)
     task *t;
     var *args;
{
  if (args->num)
    in_push(t, make_error(t, E_ARGS));
  else
    in_push(t, newnum(t->counter));
}

void bi_task_id(t, args)
     task *t;
     var *args;
{
  if (args->num)
    in_push(t, make_error(t, E_ARGS));
  else
    in_push(t, newnum(t->task_id));
}

void bi_suspend(t, args)
     task *t;
     var *args;
{
  int ct;
  static var_type types[] = { V_NUM };
  Error e;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE) {
    in_push(t, make_error(t, e));
    return;
  }

  if (args->list[0]->num < 0) {
    in_push(t, make_error(t, E_INVARG));
    return;
  }

  ct = time(0);
  t->type = TA_SUSPENDED;
  t->wakeup = ct + args->list[0]->num;
  in_push(t, NULL);
}

void bi_die(t, args)
     task *t;
     var *args;
{
  if (args && args->num) {
    in_push(t, make_error(t, E_ARGS));
    return;
  }

  t->type = TA_DYING;
  in_push(t, NULL);
}

void bi_kill_task(t, args)
     task *t;
     var *args;
{
  static var_type types[] = { V_NUM };
  Error e;
  task *temp;
  int result;

  if ((e = inv_args(args, 1, 1, types)) != E_NONE)
    in_push(t, make_error(t, e));
  else {
    temp = tasks;
    result = -1;

    while (temp) {
      if (temp->task_id == args->list[0]->num) {
	result = 0;
	temp->type = TA_DYING;
	break;
      }

      temp = temp->next;
    }
    in_push(t, newnum(result));
  }
}

void bi_queued_tasks(t, args)
     task *t;
     var *args;
{
  in_push(t, NULL);
}
