/*
 * stalk/termd.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "termd.h"
#include "parser.h"
#include "functions.h"
#include "lines.h"
#include "buffer.h"
#include "stalk.h"
#include "stream.h"
#include "stuff.h"
#include "var.h"
#include "task.h"

#ifdef SYSVTTY
static struct termio oldtermstate;
#endif
#ifdef POSIXTTY
static struct termios oldtermstate;
#endif
#ifdef BSDTTY
static struct sgttyb oldtermstate;
#endif

static char *backspace_str=NULL;
static char *space_str=NULL;
/*
 * backspace_str will store a bunch of backpace characters - and save us
 * time when we don't have to use a lot of for loops to back up more than
 * one character. space_str is similar, but allows us to overwrite text
 * with blank characters (spaces)
 */

/* each buffer must have a reference to the following */
static function_binding fb[] = {
  {"backspace", d_backspace},
  {"char_backward", d_backward},
  {"char_forward", d_forward},
  {"line_end", d_eol},
  {"line_start", d_bol},
  {"hist_backward", d_previous_line},
  {"hist_forward", d_next_line},
  {"line_kill", d_kill},
  {"forward_delete", d_ctrlD},
  {"refresh", d_redraw},
  {"save_buffer", d_save_buffer},
  {"reconnect", d_connect_new},
  {NULL, NULL}
};

void d_setup_funcs()
{
  signal(SIGINT, quit);              /* Control-C. */
  signal(SIGTSTP, handle_interrupt); /* Control-Z. */

  setup_screen = d_setup_screen;
  end_screen = d_end_screen;
  printchar = d_printchar;
  displaystr = d_displaystr;
  displaybold = d_displaystr;
  handle_return = d_return;
  print_title = do_nothing;
  backspace = d_backspace;
  backward = d_backward;
  forward = d_forward;
  eol = d_eol;
  bol = d_bol;
  previous_line = d_previous_line;
  next_line = d_next_line;
  kill_line = d_kill;
  ctrlD = d_ctrlD;
  save_kb = d_save_kb;
  restore_kb = d_restore_kb;
  redraw = d_redraw;
  page_down = do_nothing;
  page_up = do_nothing;
  scroll_down = do_nothing;
  scroll_up = do_nothing;
  save_buffer = d_save_buffer;
  connect_new = d_connect_new;
  new_buffer = d_new_buffer;
}

int d_new_buffer(prev, name, Lines)
     buffer *prev;
     char *name;
     int Lines;
{
  buffer *b;

  /* can only have one buffer in dumb terminal mode. */
  if (prev)
    return 0;

  b = newbuffer();

  b->name = NULL;
  b->global = NULL;
  b->local_bindings = NULL;
  b->function_bindings = fb;
  return 1;
}

void d_setup_screen()
{
  char bp[1024], *t;
  int x;
#ifdef SYSVTTY
  struct termio blob;

  if (ioctl(0, TCGETA, &blob) == -1)
    die("TCGETA ioctl error\n");
  ioctl(0, TCGETA, &oldtermstate);
  blob.c_lflag &= ~ECHO;
  blob.c_lflag &= ~ECHOE;
  blob.c_lflag &= ~ICANON;
  blob.c_cc[VMIN] = 0;
  blob.c_cc[VTIME] = 0;
  if (ioctl(0, TCSETAF, &blob) == -1)
    die("TCSETAF ioctl error\n");
#endif
#ifdef POSIXTTY
  /* we're using a POSIX compliant OS. */
  struct termios blob;

  if (tcgetattr(FD_IN, &blob) == -1)
    die("tcgetattr error\n");
  tcgetattr(FD_IN, &oldtermstate);
  blob.c_lflag &= ~ICANON;
  if (tcsetattr(FD_IN, TCSANOW, &blob) == -1)
    die("tcsetattr error\n");
  blob.c_lflag &= ~ECHO;
  if (tcsetattr(FD_IN, TCSADRAIN, &blob) == -1)
    die("tcsetattr error\n");
#endif
#ifdef BSDTTY
  /* we're using BSD */
  struct sgttyb blob;

  if (ioctl(FD_IN, TIOCGETP, &blob) == -1)
    die("TIOCGETP ioctl error\n");
  ioctl(FD_IN, TIOCGETP, &oldtermstate);
  blob.sg_flags |= CBREAK;
  blob.sg_flags &= ~ECHO;
  if (ioctl(FD_IN, TIOCSETP, &blob) == -1)
    die("TIOCSETP ioctl error\n");
#endif

  t = getenv("TERM");
  if (strcmp(t, "")) {
    cols = 80; /***/ /* best guess * CHANGE: use TIOCGWINSZ */
  } else {
    tgetent(bp, t);
    cols = tgetnum("co");
  }

  /* allocate backspace_str and fill it with backspaces */
  /* allocate space_str and fill it with spaces */
  if (!backspace_str) {
    backspace_str = (char *)malloc(cols+1);
    space_str = (char *)malloc(cols+1);
    for (x=cols+1; x;) {
      backspace_str[x-1] = '\010';
      space_str[--x] = ' ';
    }
  }
}

void d_end_screen()
{
#ifdef SYSVTTY
  ioctl(0, TCSETAF, &oldtermstate);
#endif
#ifdef POSIXTTY
  tcsetattr(FD_IN, TCSANOW, &oldtermstate);
#endif
#ifdef BSDTTY
  ioctl(FD_IN, TIOCSETP, &oldtermstate);
#endif
}

/*
 * printchar() will send the character to the screen, update the
 * keyboard buffer, and move the text over that appears after the
 * character.
 */
#ifdef __STDC__
void d_printchar(buffer *buf, char c)
#else
void d_printchar(buf, c)
     buffer *buf;
     char c;
#endif
{
  int x, move;

  for(x=buf->keyboard.len++; x>=buf->keyboard.pos; x--)
    buf->keyboard.line[x+1] = buf->keyboard.line[x];
  buf->keyboard.line[buf->keyboard.pos++] = c;

  write(FD_OUT, &c, 1);
  if (buf->keyboard.len > buf->keyboard.pos) {
    if (buf->keyboard.pos/cols != buf->keyboard.len/cols)
      move=cols - buf->keyboard.pos%cols - 1;
    else
      move=buf->keyboard.len-buf->keyboard.pos;
    write(FD_OUT, buf->keyboard.line+buf->keyboard.pos, move);
    write(FD_OUT, backspace_str, move);
  }
}

/*
 * displaystr() will send the string to the screen (wrapping it),
 * and put the string in the line queue
 */
void d_displaystr(buf, s)
     buffer *buf;
     char *s;
{
  static char *tmp=NULL;
  int x, displayed, len, i, j;
  keyboard_state save;

  if (!buf) {
    fprintf(stderr, "%s\n", s);
    return;
  }

  d_save_kb(buf, &save);

  if (!tmp)
    tmp = (char *)malloc(sizeof(char) * (cols+5));

  len = strlen(s);
  displayed = 0;
  
  while((displayed != len) || (len == 0)) {
    /* find the last space before the end of the screen */
    if (len <= displayed+cols-1)
      x=len;
    else {
      x=cols+displayed-1; 
      while(x>displayed && s[x]!=' ')
	x--;
    }
    
    if (x==displayed) /* word longer than the screen. hmmm... split it */
      x=(cols+displayed)<len ? (cols+displayed-1) : len;
    
    /* copy the string into the temporary buffer, omiting any control chars */
    for (i=displayed, j=0; i<x; i++)
      if ((isprint(s[i]) || isspace(s[i])) && (s[i] >= ' '))
        tmp[j++] = s[i];
    tmp[j]='\0';
    
    /*
     * put the space at the end of the line in the line queue
     * this way it doesn't get printed to the screen - avoiding
     * annoying extra lines
     */
    if ((x<len) && (isspace(s[x]))) x++;
    displayed = x;
    
    put_line_in_linequeue(buf, tmp, 0);
    
    printf("%s\n\r", tmp);
    
    if (len == 0)
      displayed = len = 1;
  }

  d_restore_kb(buf, &save);
}

/*
 * return() will handle carriage returns - it will put the
 * cursor on the next line, add the line to the history list, update
 * the line queue, and send the line to any open socket
 */
void d_return(buf)
     buffer *buf;
{
  statement *temp;
  char string[9999];
  int x;

  d_eol(buf); /* go to the end of the line so that return
		      doesn't chop off part of the line */
  printf("\n\r");

  buf->keyboard.line[buf->keyboard.len] = '\0';
  put_line_in_history(buf, buf->keyboard.line);
  put_line_in_linequeue(buf, buf->keyboard.line, 1);

  if (buf->streams->on_input &&
      func_exists(global_funcs, buf->streams->on_input)) {
    var *v;
    program *p;

    v = newlist(1);
    v->list[0] = newstring(mystrdup(buf->keyboard.line));
    buf->keyboard.len = buf->keyboard.pos = 0;
    buf->history_pos = -1;
    p = find_program(global_funcs, buf->streams->on_input);
    push_task(buf, p, v);
    var_free(v);
  } else {
    if ((buf->keyboard.len > 1) && (buf->keyboard.line[0] == '/')) {
      buf->keyboard.line[buf->keyboard.len++] = '\n';
      buf->keyboard.line[buf->keyboard.len] = '\0';
      for (x=0; x<buf->keyboard.len-1; x++)
	string[x] = buf->keyboard.line[x+1];
      string[buf->keyboard.len-1] = '\0';
      buf->keyboard.len = buf->keyboard.pos = 0;
      buf->history_pos = -1;
      temp = my_parse(buf, string);
      if (temp) {
	program *p;

	p = NEW(program);
	p->base = temp;
	p->ref = 0;
	p->evaled = 1;
	p->debug = 1;
	p->name = mystrdup("* keyboard program *");
	push_task(buf, p, NULL);
      }
    } else {
      buf->keyboard.line[buf->keyboard.len++] = '\r';
      buf->keyboard.line[buf->keyboard.len++] = '\n';
      buf->keyboard.line[buf->keyboard.len] = '\0';
      send_to_streams(buf, ALL_TASKS, buf->keyboard.line);
      buf->keyboard.len = buf->keyboard.pos = 0;
      buf->history_pos = -1;
    }
  }

  while (buf->history->cursor->next != NULL)
    buf->history->cursor = buf->history->cursor->next;
  buf->history->pos = buf->history->size;
}

void d_backspace(buf)
     buffer *buf;
{
  int x, chars, space;

  if (buf->keyboard.pos == 0) {
    beep(buf); return;
  }

  if (buf->keyboard.pos % cols) {
    if (buf->keyboard.len <= buf->keyboard.pos) {
      write(FD_OUT, "\010 \010", 3);
    } else {
      write(FD_OUT, "\010", 1);
      /* redraw to the end of line */
      if ((buf->keyboard.len/cols) != (buf->keyboard.pos/cols)) {
	chars = cols - buf->keyboard.pos%cols;
	space = 0;
      } else {
        chars = buf->keyboard.len - buf->keyboard.pos;
	space = 1;
      }
      write(FD_OUT, buf->keyboard.line+buf->keyboard.pos, chars);
      if (space)
	write(FD_OUT, " ", 1);
      write(FD_OUT, backspace_str, chars+space);
    }
  } else {
    /* redraw the previous line, leaving the cursor at the end */
    write(FD_OUT, buf->keyboard.line+buf->keyboard.pos-cols, cols-1);
  }

  for(x=buf->keyboard.pos-1; x<buf->keyboard.len-1; x++)
    buf->keyboard.line[x] = buf->keyboard.line[x+1];

  buf->keyboard.pos--; buf->keyboard.len--;
}

void d_backward(buf)
     buffer *buf;
{
  if (buf->keyboard.pos == 0) {
    beep(buf);
    return;
  }

  if (!((buf->keyboard.pos) % cols)) {
    write(FD_OUT, backspace_str, cols-1);
    write(FD_OUT, buf->keyboard.line+buf->keyboard.pos-cols, cols-1);
  } else
    write(FD_OUT, "\010", 1);

  buf->keyboard.pos--;
}

void d_forward(buf)
     buffer *buf;
{
  int redraw_lines;

  if (buf->keyboard.pos >= buf->keyboard.len) {
    beep(buf);
    return;
  }

  buf->keyboard.pos++;

  if (!((buf->keyboard.pos) % cols)) {
    write(FD_OUT, backspace_str, cols-1);
    redraw_lines = ((buf->keyboard.len / cols) == (buf->keyboard.pos / cols))
      ? buf->keyboard.len - buf->keyboard.pos : cols-1;
    write(FD_OUT, buf->keyboard.line+buf->keyboard.pos, redraw_lines);
    write(FD_OUT, space_str, cols-1-redraw_lines);
    write(FD_OUT, backspace_str, cols-1);
  } else
    write(FD_OUT, buf->keyboard.line+buf->keyboard.pos-1, 1);
}

void d_eol(buf)
     buffer *buf;
{
  write(FD_OUT, buf->keyboard.line+buf->keyboard.pos,
	buf->keyboard.len-buf->keyboard.pos);
  buf->keyboard.pos = buf->keyboard.len;
}

void d_bol(buf)
     buffer *buf;
{
  write(FD_OUT, backspace_str, buf->keyboard.pos%cols);
  if (buf->keyboard.pos >= cols) {
    write(FD_OUT, buf->keyboard.line, cols-1);
    write(FD_OUT, backspace_str, cols-1);
  }
  buf->keyboard.pos = 0;
}

void d_previous_line(buf)
     buffer *buf;
{
  int x;

  if (buf->history->size == 0 || buf->history_pos == 1)
    return;

  if (buf->history_pos == -1)
    buf->history_pos = buf->history->size;
  else
    buf->history_pos--;
  while (buf->history->pos > buf->history_pos)
    retreat_lq(buf->history);

  /* erase old line, put new line in buffer, and print it */
  write(FD_OUT, backspace_str, buf->keyboard.pos%cols);
  write(FD_OUT, space_str, cols-1);
  write(FD_OUT, backspace_str, cols-1);
  for(x = buf->keyboard.len = buf->keyboard.pos
      = strlen(buf->history->cursor->string); x; x--)
    buf->keyboard.line[x-1] = buf->history->cursor->string[x-1];
  write(FD_OUT, buf->keyboard.line, buf->keyboard.len);
}

void d_next_line(buf)
     buffer *buf;
{
  int x;

  if (buf->history_pos == -1)
    return;
  if (buf->history_pos == buf->history->size) {
    buf->history_pos = -1;
    d_bol(buf);
    d_kill(buf);
    return;
  } else
    buf->history_pos++;
  while ((buf->history->pos < buf->history_pos) && (buf->history_pos != -1))
    advance_lq(buf->history);

  /* erase old line, put new line in buffer, and print it */
  write(FD_OUT, backspace_str, buf->keyboard.pos%cols);
  write(FD_OUT, space_str, cols-1);
  write(FD_OUT, backspace_str, cols-1);
  for(x = buf->keyboard.len = buf->keyboard.pos
      = strlen(buf->history->cursor->string); x; x--)
    buf->keyboard.line[x-1] = buf->history->cursor->string[x-1];
  write(FD_OUT, buf->keyboard.line, buf->keyboard.len);
}

void d_kill(buf)
     buffer *buf;
{
  int erase; /* number of characters to erase */

  if (buf->keyboard.pos/cols != buf->keyboard.len/cols)
    erase=cols - buf->keyboard.pos%cols - 1;
  else
    erase=buf->keyboard.len-buf->keyboard.pos;

  write(FD_OUT, space_str, erase);
  write(FD_OUT, backspace_str, erase);

  buf->keyboard.len = buf->keyboard.pos;
}

void d_ctrlD(buf)
     buffer *buf;
{
  int x;

  if (buf->keyboard.pos >= buf->keyboard.len) {
    beep(buf);
    return;
  }

  if ((buf->keyboard.pos+1) % cols) {
    d_forward(buf);
    d_backspace(buf);
  } else {
    /* can't do the above trick, since it'll draw the next line of text */
    buf->keyboard.len--;
    for (x=buf->keyboard.pos; x<buf->keyboard.len; x++)
      buf->keyboard.line[x] = buf->keyboard.line[x+1];
  }
}

void d_save_kb(buf, save)
     buffer *buf;
     keyboard_state *save;
{
  int erase;

  save->len = buf->keyboard.len;
  save->pos = buf->keyboard.pos;

  if (!save->len)
    return;

  write(FD_OUT, backspace_str, save->pos%cols);

  if ((save->len/cols) != (save->pos/cols))
    erase = cols-1;
  else
    erase=save->len%cols;

  write(FD_OUT, space_str, erase);
  write(FD_OUT, backspace_str, erase);

  buf->keyboard.len = 0;
  buf->keyboard.pos = 0;
}

void d_restore_kb(buf, save)
     buffer *buf;
     keyboard_state *save;
{
  if (!save->len)
    return;

  buf->keyboard.len = save->len;
  buf->keyboard.pos = save->pos;

  if ((save->len/cols) != (save->pos/cols)) {
    write(FD_OUT, buf->keyboard.line+(save->pos/cols)*cols, cols-1);
    write(FD_OUT, backspace_str, cols-save->pos%cols-1);
  } else {
    write(FD_OUT, buf->keyboard.line+(save->pos/cols)*cols, save->len%cols);
    write(FD_OUT, backspace_str, save->len - save->pos);
  }
}

void d_redraw(buf)
     buffer *buf;
{
  write(FD_OUT, "\n\r", 2);
  write(FD_OUT, buf->keyboard.line, buf->keyboard.len);
  if ((buf->keyboard.len / cols) == (buf->keyboard.pos / cols))
    write(FD_OUT, backspace_str, buf->keyboard.len - buf->keyboard.pos);
  else {
    write(FD_OUT, "\n\r", 2);
    write(FD_OUT, buf->keyboard.line+(buf->keyboard.pos/cols)*cols, cols-1);
    write(FD_OUT, backspace_str, cols-buf->keyboard.pos%cols-1);
  }
}

void d_save_buffer(buf)
     buffer *buf;
{
  line *lp;
  FILE *fp;

  lp = buf->lines->cursor;
  while (lp->prev != NULL)
    lp = lp->prev;
  fp = fopen("stalk.log", "w");
  while (lp->next != NULL) {
    fputs(lp->next->string, fp);
    if (lp->next->eol)
      fprintf(fp, "\n");
    lp = lp->next;
  }
  fclose(fp);
}

void d_connect_new(buf)
     buffer *buf;
{
  char name[1024];
  int port;
  char s[100];

  buf->keyboard.line[buf->keyboard.len] = '\0';
  sscanf(buf->keyboard.line, "%s %d", name, &port);

  d_bol(buf);
  d_kill(buf);

  sprintf(s, "connecting to %s %d", name, port);
  d_displaystr(buf, "");
  d_displaystr(buf, "---");
  d_displaystr(buf, "");
  d_displaystr(buf, s);
  d_displaystr(buf, "");
  d_displaystr(buf, "---");
  d_displaystr(buf, "");

  new_cstream(buf, name, port);
}
