/*
 * stalk/termt.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* abandon hope all ye who enter here */

#include "include.h"
#include "termt.h"
#include "parser.h"
#include "functions.h"
#include "lines.h"
#include "buffer.h"
#include "stalk.h"
#include "stream.h"
#include "stuff.h"
#include "var.h"
#include "task.h"

#ifdef SYSVTTY
static struct termio oldtermstate;
#endif
#ifdef POSIXTTY
static struct termios oldtermstate;
#endif
#ifdef BSDTTY
static struct sgttyb oldtermstate;
#endif

static char tbuffer[1024], termentry[1024];
static char *boldon, *boldoff, *clear, *move, *clr_eol, *del_char;
static char *insert_on, *insert_off, *start_tc, *end_tc;
static char *set_scroll_region, *scroll_forward, *scroll_reverse;
static char *revvid_on, *revvid_off;
static char *tempptr, *term, *S;
static char *spaces;
#ifndef hpux
#if defined(linux)
extern speed_t ospeed;
#else
extern short ospeed;
#endif /* linux */
extern char PC;
#endif /* !hpux */

#define update_cursor(buf) \
  move_cursor((buf)->keyboard.pos%cols, \
              (buf)->row+(buf)->keyboard.pos/cols)
#define move_cursor(x, y) \
  S=tgoto(move, (x), (y)), \
  TPUTS(S), \
  fflush(stdout)
#define TPUTS(s) tputs((s), 1, (putchar))
#define ROW(x) ((x)/cols)

/* each buffer must have a reference to the following */
static function_binding fb[] = {
  {"backspace", t_backspace},
  {"char_backward", t_backward},
  {"char_forward", t_forward},
  {"line_end", t_eol},
  {"line_start", t_bol},
  {"hist_backward", t_previous_line},
  {"hist_forward", t_next_line},
  {"line_kill", t_kill},
  {"forward_delete", t_ctrlD},
  {"refresh", refresh},
  {"save_buffer", t_save_buffer},
  {"reconnect", t_connect_new},
  {"scroll_down_page", t_page_down},
  {"scroll_up_page", t_page_up},
  {"scroll_down_line", t_scroll_down},
  {"scroll_up_line", t_scroll_up},
  {"next_buffer", next_buffer},
  {NULL, NULL}
};

void t_setup_funcs()
{
  signal(SIGINT, quit);              /* Control-C. */
  signal(SIGTSTP, handle_interrupt); /* Control-Z. */

  setup_screen = t_setup_screen;
  end_screen = t_end_screen;
  printchar = t_printchar;
  displaystr = t_displaystr;
  displaybold = t_displaybold;
  handle_return = t_return;
  print_title = t_print_title;
  backspace = t_backspace;
  backward = t_backward;
  forward = t_forward;
  eol = t_eol;
  bol = t_bol;
  previous_line = t_previous_line;
  next_line = t_next_line;
  kill_line = t_kill;
  ctrlD = t_ctrlD;
  save_kb = t_save_kb;
  restore_kb = t_restore_kb;
  redraw = t_redraw;
  page_down = t_page_down;
  page_up = t_page_up;
  scroll_down = t_scroll_down;
  scroll_up = t_scroll_up;
  save_buffer = t_save_buffer;
  connect_new = t_connect_new;
  new_buffer = t_new_buffer;
}

/*
 * lines is actually a request for the number of lines to split by, with the
 * following convention:
 * lines negative: give the old buffer (bottom) a number of lines
 *   equal to the opposite of this number
 * lines positive: give the new buffer (top) this many lines
 * lines 0: split evenly
 */
int t_new_buffer(prev, name, lines)
     buffer *prev;
     char *name;
     int lines;
{
  buffer *b;

  if (prev && (prev->height < 7))
    return 0;

  b = newbuffer();

  b->global = NULL;
  b->connected = 0;
  b->scrollmode = 0;
  b->name = mystrdup(name);
  if (prev) {
    b->startline = prev->startline;
    if (lines < 0) {
      b->height = MIN(MAX(2, prev->height-1+lines), prev->height-3);
    } else if (!lines) {
      b->height = (prev->height-1)/2;
    } else {
      b->height = MIN(MAX(2, prev->height-1-lines), prev->height-3);
    }
    b->row = prev->startline;
    prev->startline = b->startline+b->height+1;
    prev->height = prev->height-1-b->height;
    prev->row = MIN(prev->startline+prev->row-b->startline,
		    prev->startline+prev->height-2-prev->keyboard.len/cols);
  } else {
    b->startline = 0;
    b->height = rows-1;
    b->row = 0;
  }
  b->local_bindings = NULL;
  b->function_bindings = fb;

  t_redraw(b);
  if (prev)
    t_redraw(prev);

  return b->buf_id;
}

void t_setup_screen()
{
  int x;

#ifdef SYSVTTY
  struct termio blob;

  if (ioctl(0, TCGETA, &blob) == -1)
    die("TCGETA ioctl error\n");
  ioctl(0, TCGETA, &oldtermstate);
  blob.c_lflag &= ~ECHO;
  blob.c_lflag &= ~ECHOE;
  blob.c_lflag &= ~ICANON;
  blob.c_lflag &= ~TAB3;   /* turn off tab expansion */
  blob.c_cc[VMIN] = 0;
  blob.c_cc[VTIME] = 0;
  if (ioctl(0, TCSETAF, &blob) == -1)
    die("TCSETAF ioctl error\n");
#endif
#ifdef POSIXTTY
  /* we're using a POSIX compliant OS */
  struct termios blob;

  if (tcgetattr(FD_IN, &blob) == -1)
    die("tcgetattr error\n");
  tcgetattr(FD_IN, &oldtermstate);
  ospeed = cfgetospeed(&blob);
  blob.c_lflag &= ~ICANON; /* turn on character mode instead of line mode */
  blob.c_lflag &= ~TAB3;   /* turn off tab expansion */
  if (tcsetattr(FD_IN, TCSANOW, &blob) == -1)
    die("tcsetattr error\n");
  blob.c_lflag &= ~ECHO;   /* turn off echo. fun! :-| */
  if (tcsetattr(FD_IN, TCSADRAIN, &blob) == -1)
    die("tcsetattr error\n");
#endif
#ifdef BSDTTY
  /* we're using BSD */
  struct sgttyb blob;

  if (ioctl(FD_IN, TIOCGETP, &blob) == -1)
    die("TIOCGETP ioctl error\n");
  ioctl(FD_IN, TIOCGETP, &oldtermstate);
  blob.sg_flags &= ~XTABS;  /* turn off tab expansion */
  blob.sg_flags |= CBREAK;  /* turn on character mode instead of line mode */
  blob.sg_flags &= ~ECHO;   /* turn off echo. fun! :-| */
  ospeed = blob.sg_ospeed;  /* output speed - used to see how much padding
			     * the terminal needs */
  if (ioctl(FD_IN, TIOCSETP, &blob) == -1)
    die("TIOCSETP ioctl error\n");
#endif

  if (!(term = getenv("TERM")))
    die("term type not found!\n");

  tgetent(termentry, term);
  cols = tgetnum("co");
  rows = tgetnum("li");
  
  tempptr = tbuffer;
  if ((boldon = tgetstr("md", &tempptr)))
    boldoff = tgetstr("me", &tempptr);
  else if ((boldon = tgetstr("so", &tempptr)))
    boldoff = tgetstr("se", &tempptr);
  clear = tgetstr("cl", &tempptr);
  clr_eol = tgetstr("ce", &tempptr);
  move = tgetstr("cm", &tempptr);
  del_char = tgetstr("dc", &tempptr);
  insert_on = tgetstr("im", &tempptr);
  insert_off = tgetstr("ei", &tempptr);
  start_tc = tgetstr("ti", &tempptr);
  end_tc = tgetstr("te", &tempptr);
  set_scroll_region = tgetstr("cs", &tempptr);
  scroll_forward = tgetstr("sf", &tempptr);
  scroll_reverse = tgetstr("sr", &tempptr);
  revvid_on = tgetstr("mr", &tempptr);
  revvid_off = tgetstr("me", &tempptr);
  if (!revvid_on || !revvid_off) {
    revvid_on = boldon;
    revvid_on = boldoff;
  }

#ifndef hpux
  /* hpux doesn't have a padding character, it seems */
  if ((S = tgetstr("pc", &tempptr)))
    PC = *S; /* padding character */
#endif

  TPUTS(start_tc);
  TPUTS(clear);
  fflush(stdout);

  spaces = (char *)malloc(cols);
  for (x=cols; x; x--)
    spaces[x-1] = ' ';
}

void t_end_screen()
{
#ifdef SYSVTTY
  ioctl(0, TCSETAF, &oldtermstate);
#endif
#ifdef POSIXTTY
  tcsetattr(FD_IN, TCSANOW, &oldtermstate);
#endif
#ifdef BSDTTY
  ioctl(FD_IN, TIOCSETP, &oldtermstate);
#endif

  TPUTS(clear);
  TPUTS(end_tc);
  fflush(stdout);
}

void t_boldon()
{
  if (!boldon || !boldoff)
    return;

  TPUTS(boldon);
  fflush(stdout);
}

void t_boldoff()
{
  if (!boldon || !boldoff)
    return;

  TPUTS(boldoff);
  fflush(stdout);
}

/*
 * printchar() will send the character to the screen, update the
 * keyboard buffer, and move the text over that appears after the
 * character.
 */
#ifdef __STDC__
void t_printchar(buffer *buf, char c)
#else
void t_printchar(buf, c)
     buffer *buf;
     char c;
#endif
{
  int x;

  if (buf->scrollmode)
    t_unscroll(buf);

  /* check to see if we're at the end of the buffer and need to scroll */
  if ((buf->keyboard.len%cols == cols-1) &&
      (buf->row + buf->keyboard.len/cols == buf->height+buf->startline-1)) {
    t_scroll_forward_one(buf);
    update_cursor(buf);
  }

  if (buf->keyboard.len == buf->keyboard.pos) {
    putchar(c);
    buf->keyboard.line[buf->keyboard.len++] = c;
    buf->keyboard.pos++;
  } else {
    for(x=buf->keyboard.len++; x>=buf->keyboard.pos; x--)
      buf->keyboard.line[x+1] = buf->keyboard.line[x];
    buf->keyboard.line[buf->keyboard.pos++] = c;
    if (insert_on && insert_off) {
      TPUTS(insert_on);
      putchar(c);
      TPUTS(insert_off);

      for(x=ROW(buf->keyboard.pos); x<ROW(buf->keyboard.len); x++) {
	move_cursor(0, buf->row+x+1);
	TPUTS(insert_on);
	putchar(buf->keyboard.line[(x+buf->keyboard.pos/cols+1)*cols]);
	TPUTS(insert_off);
      }
    } else {
      fflush(stdout);
      write(FD_OUT, buf->keyboard.line+buf->keyboard.pos-1,
	    buf->keyboard.len-buf->keyboard.pos+1);
    }
    update_cursor(buf);
  }
  fflush(stdout);
}

/*
 * displaystr() will send the string to the screen (wrapping it),
 * and put the string in the line queue
 */
void t_displaystr(buf, str)
     buffer *buf;
     char *str;
{
  static char *tmp=NULL;
  int x, displayed, len, i, j;
  keyboard_state save;

  if (!buf) {
    fprintf(stderr, "%s\n", str);
    return;
  }

  t_save_kb(buf, &save);

  if (tmp==0)
    tmp = (char *)malloc(sizeof(char)*(cols+5));

  len = strlen(str);
  displayed = 0;

  while((displayed != len) || (len == 0)) {
    /* find the last space before the end of the screen */
    if (len <= displayed+cols-1)
      x=len;
    else {
      x=cols+displayed-1; 
      while(x>displayed && str[x]!=' ')
	x--;
    }

    if (x==displayed) /* word longer than the screen. hmmm... split it */
      x = MIN(cols+displayed-1, len);

    /* copy the string into the temporary buffer, omiting any control chars */
    for (i=displayed, j=0; i<x; i++)
      if ((isprint(str[i]) || isspace(str[i])) && (str[i] >= ' '))
	tmp[j++] = str[i];
    tmp[j]='\0';

    /*
     * put the space at the end of the line in the line queue
     * this way it doesn't get printed to the screen - avoiding
     * annoying extra lines
     */
    while ((x<len) && (isspace(str[x])))
      x++;
    displayed = x;

    if (!buf->scrollmode) {
      if (buf->row == buf->height+buf->startline-1)
	t_scroll_forward_one(buf);
      update_cursor(buf);
      printf("%s\n\r", tmp);
      buf->row++;
      update_cursor(buf);
    }
    
    put_line_in_linequeue(buf, tmp, !buf->scrollmode);

    if (len == 0)
      displayed = len = 1;
  }

  t_restore_kb(buf, &save);
}

void t_displaybold(buf, str)
     buffer *buf;
     char *str;
{
  t_boldon();
  t_displaystr(buf, str);
  t_boldoff();
}

/*
 * return() will handle carriage returns - it will put the
 * cursor on the next line, add the line to the history list, update
 * the line queue, and send the line to any open socket
 */
void t_return(buf)
     buffer *buf;
{
  statement *temp;
  char string[9999];
  int x;

  /*
   * go to the end of the line so that return
   * doesn't chop off part of the line
   */
  t_eol(buf);

  if (buf->row+buf->keyboard.len/cols == buf->height+buf->startline-1)
    t_scroll_forward_one(buf);

  buf->row += buf->keyboard.len/cols+1;

  buf->keyboard.line[buf->keyboard.len] = '\0';
  put_line_in_history(buf, buf->keyboard.line);
  put_line_in_linequeue(buf, buf->keyboard.line, 1);

  if (buf->streams->on_input &&
      func_exists(global_funcs, buf->streams->on_input)) {
    var *v;
    program *p;

    v = newlist(1);
    v->list[0] = newstring(mystrdup(buf->keyboard.line));
    buf->keyboard.len = buf->keyboard.pos = 0;
    buf->history_pos = -1;
    update_cursor(buf);
    p = find_program(global_funcs, buf->streams->on_input);
    push_task(buf, p, v);
    var_free(v);
  } else {
    if ((buf->keyboard.len > 1) && (buf->keyboard.line[0] == '/')) {
      buf->keyboard.line[buf->keyboard.len++] = '\n';
      buf->keyboard.line[buf->keyboard.len] = '\0';
      for (x=0; x<buf->keyboard.len-1; x++)
	string[x] = buf->keyboard.line[x+1];
      string[buf->keyboard.len-1] = '\0';
      buf->keyboard.len = buf->keyboard.pos = 0;
      buf->history_pos = -1;
      update_cursor(buf);

      temp = my_parse(buf, string);
      if (temp) {
	program *p;

	p = NEW(program);
	p->base = temp;
	p->ref = 0;
	p->evaled = 1;
	p->debug = 1;
	p->name = mystrdup("* keyboard program *");
	push_task(buf, p, NULL);
      }
    } else {
      buf->keyboard.line[buf->keyboard.len++] = '\r';
      buf->keyboard.line[buf->keyboard.len++] = '\n';
      buf->keyboard.line[buf->keyboard.len] = '\0';
      send_to_streams(buf, ALL_TASKS, buf->keyboard.line);
      buf->keyboard.len = buf->keyboard.pos = 0;
      buf->history_pos = -1;
      update_cursor(buf);
    }
  }

  while (buf->history->cursor->next)
    buf->history->cursor = buf->history->cursor->next;
  buf->history->pos = buf->history->size;
}

/*
 * print_title() will print the title of the buffer below it in reverse video
 */
void t_print_title(buf)
     buffer *buf;
{
  move_cursor(0, buf->startline + buf->height);
  if (revvid_on && revvid_off)
    TPUTS(revvid_on), fflush(stdout);
  write(FD_OUT, buf->name, MIN(strlen(buf->name), cols));
  write(FD_OUT, spaces, cols-strlen(buf->name));
  if (revvid_on && revvid_off)
    TPUTS(revvid_off);
  update_cursor(get_buffer(current_buffer));
}

void t_backspace(buf)
     buffer *buf;
{
  int x;

  if (buf->scrollmode)
    t_unscroll(buf);

  if (buf->keyboard.pos == 0) {
    beep(buf);
    return;
  }

  --buf->keyboard.len;
  --buf->keyboard.pos;
  for(x=buf->keyboard.pos; x<buf->keyboard.len; x++)
    buf->keyboard.line[x] = buf->keyboard.line[x+1];
  update_cursor(buf);
  if (del_char) {
    TPUTS(del_char);
    fflush(stdout);
    if (buf->keyboard.len > buf->keyboard.pos) {
      for (x=ROW(buf->keyboard.pos+1); x<ROW(buf->keyboard.len+1); x++) {
	move_cursor(cols-1, buf->row+x);
	putchar((x+1)*cols-1 < buf->keyboard.len
		? buf->keyboard.line[(x+1)*cols-1] : ' ');
	move_cursor(0, buf->row+x+1);
	TPUTS(del_char);
	fflush(stdout);
      }
    }
  } else {
    write(FD_OUT, buf->keyboard.line+buf->keyboard.pos,
	  buf->keyboard.len-buf->keyboard.pos);
    write(FD_OUT, " ", 1);
  }
  update_cursor(buf);
}

void t_backward(buf)
     buffer *buf;
{
  if (buf->scrollmode)
    t_unscroll(buf);

  if (buf->keyboard.pos == 0) {
    beep(buf);
    return;
  }

  buf->keyboard.pos--;
  update_cursor(buf);
}

void t_forward(buf)
     buffer *buf;
{
  if (buf->scrollmode)
    t_unscroll(buf);

  if (buf->keyboard.pos>=buf->keyboard.len) {
    beep(buf);
    return;
  }

  buf->keyboard.pos++;
  update_cursor(buf);
}

void t_eol(buf)
     buffer *buf;
{
  if (buf->scrollmode)
    t_unscroll(buf);

  buf->keyboard.pos = buf->keyboard.len;
  update_cursor(buf);
}

void t_bol(buf)
     buffer *buf;
{
  if (buf->scrollmode)
    t_unscroll(buf);

  buf->keyboard.pos = 0;
  update_cursor(buf);
}

void t_previous_line(buf)
     buffer *buf;
{
  int x;

  if (buf->scrollmode)
    t_unscroll(buf);

  if (buf->history->size == 0 || buf->history_pos == 1)
    return;

  if (buf->history_pos == -1)
    buf->history_pos = buf->history->size;
  else
    buf->history_pos--;
  while (buf->history->pos > buf->history_pos)
    retreat_lq(buf->history);
  t_bol(buf);
  t_kill(buf);
  for(x=0; x<strlen(buf->history->cursor->string); x++)
    t_printchar(buf, buf->history->cursor->string[x]);
}

void t_next_line(buf)
     buffer *buf;
{
  int x;

  if (buf->scrollmode)
    t_unscroll(buf);

  if (buf->history_pos == -1)
    return;
  if (buf->history_pos == buf->history->size)
    buf->history_pos = -1;
  else
    buf->history_pos++;
  while ((buf->history->pos < buf->history_pos) && (buf->history_pos != -1))
    advance_lq(buf->history);
  t_bol(buf);
  t_kill(buf);
  if (buf->history_pos>=0)
    for(x=0; x<strlen(buf->history->cursor->string); x++)
      t_printchar(buf, buf->history->cursor->string[x]);
}

void t_kill(buf)
     buffer *buf;
{
  int x;

  if (buf->scrollmode)
    t_unscroll(buf);

  TPUTS(clr_eol);
  for (x=ROW(buf->keyboard.pos)+1; x<=ROW(buf->keyboard.len); x++) {
    move_cursor(0, buf->row+x);
    TPUTS(clr_eol);
  }
  buf->keyboard.len = buf->keyboard.pos;
  update_cursor(buf);
}

void t_ctrlD(buf)
     buffer *buf;
{
  if (buf->scrollmode)
    t_unscroll(buf);

  if (buf->keyboard.pos >= buf->keyboard.len)
    return;
  t_forward(buf);
  t_backspace(buf);
}

/*
 * this takes advantage of the fact that t_bol and t_kill do not change the
 * character buffer, only the screen and the len and pos fields
 */
void t_save_kb(buf, save)
     buffer *buf;
     keyboard_state *save;
{
  if (buf->scrollmode)
    return;

  save->len = buf->keyboard.len;
  save->pos = buf->keyboard.pos;

  t_bol(buf);
  t_kill(buf);
}

/*
 * this takes advantage of the fact that nothing will change the
 * stored keyboard line after t_save_kb is called. make sure nothing does.
 * or else.
 */
void t_restore_kb(buf, save)
     buffer *buf;
     keyboard_state *save;
{
  if (buf->scrollmode)
    return;

  update_cursor(buf);

  buf->keyboard.len = save->len;
  buf->keyboard.pos = save->pos;

  write(FD_OUT, buf->keyboard.line, buf->keyboard.len);

  update_cursor(get_buffer(current_buffer));
}

void t_redraw(buf)
     buffer *buf;
{
  int x, y;
  line *lp;

  t_print_title(buf);

  if (buf->lines->size) {
    lp = buf->lines->cursor;
    for (x=buf->row-buf->startline; x; x--)
      lp = lp->prev;

    for (x=0; lp->next && (x < buf->height); x++) {
      move_cursor(0, x+buf->startline);
      lp = lp->next;
      y = strlen(lp->string);
      write(FD_OUT, lp->string, y);
      if (y<cols)
	TPUTS(clr_eol);
      fflush(stdout);
    }
  } else
    x=0;

  if (x<buf->height) {
    move_cursor(0, x + buf->startline);
    write(FD_OUT, buf->keyboard.line,
	  MIN(buf->keyboard.len, (buf->height-x)*cols));
    TPUTS(clr_eol);
    x += buf->keyboard.len/cols+1;
  }

  for (; x<buf->height; x++) {
    move_cursor(0, x+buf->startline);
    TPUTS(clr_eol);
  }
  fflush(stdout);

  update_cursor(get_buffer(current_buffer));
}

void t_page_down(buf)
     buffer *buf;
{
  int x;

  buf->scrollmode=1;
  for (x=0; (x<buf->height/2+1) && (buf->lines->pos != buf->lines->size); x++)
    advance_lq(buf->lines);

  t_redraw(buf);
}

void t_page_up(buf)
     buffer *buf;
{
  int x;

  buf->scrollmode=1;
  for (x=0; (x<buf->height/2+1) && (buf->lines->pos > buf->row-buf->startline);
       x++)
    retreat_lq(buf->lines);

  t_redraw(buf);
}

void t_scroll_down(buf)
     buffer *buf;
{
  buf->scrollmode=1;

  if (buf->lines->pos != buf->lines->size)
    advance_lq(buf->lines);

  t_redraw(buf);
}

void t_scroll_up(buf)
     buffer *buf;
{
  buf->scrollmode=1;

  if (buf->lines->pos > buf->row-buf->startline)
    retreat_lq(buf->lines);

  t_redraw(buf);
}

void t_unscroll(buf)
     buffer *buf;
{
  buf->scrollmode = 0;
  while(buf->lines->pos!=buf->lines->size)
    advance_lq(buf->lines);

  t_redraw(buf);
}

void t_save_buffer(buf)
     buffer *buf;
{
  line *lp;
  FILE *fp;

  lp = buf->lines->cursor;
  while (lp->prev)
    lp = lp->prev;
  fp = fopen("stalk.log", "w");
  while (lp->next) {
    fputs(lp->next->string, fp);
    if (lp->next->eol)
      fprintf(fp, "\n");
    lp = lp->next;
  }
  fclose(fp);
}

void t_connect_new(buf)
     buffer *buf;
{
  char name[1024];
  int port;
  char s[100];

  buf->keyboard.line[buf->keyboard.len] = '\0';
  sscanf(buf->keyboard.line, "%s %d", name, &port);

  t_bol(buf);
  t_kill(buf);

  sprintf(s, "connecting to %s %d", name, port);
  t_displaystr(buf, "");
  t_displaystr(buf, "---");
  t_displaystr(buf, "");
  t_displaystr(buf, s);
  t_displaystr(buf, "");
  t_displaystr(buf, "---");
  t_displaystr(buf, "");

  new_cstream(buf, name, port);
}

/*
 * t_scroll_forward_one should keep buf->row up to date.
 *
 * in the case of screen without scrolling regions or line insert
 * mode, this will actually scroll half the screen instead of one line
 * this will keep people on 9600 baud (or 2400 baud - eeek) lines from
 * having to wait for scrolling too long
 */
void t_scroll_forward_one(buf)
     buffer *buf;
{
  /* scroll the screen up by one */
  if (set_scroll_region && scroll_forward) {
    S = my_tsub(set_scroll_region, buf->startline,
		buf->height+buf->startline-1);
    TPUTS(S);
    free(S);
    move_cursor(0, buf->height+buf->startline-1);
    TPUTS(scroll_forward);
    S = my_tsub(set_scroll_region, 0, rows-1);
    TPUTS(S);
    free(S);
    update_cursor(buf);
    buf->row--;
  } else {
    buf->row -= buf->height/2;
    if (buf->row < 0)
      buf->row = buf->height - buf->keyboard.len/cols - 2;
    t_redraw(buf);
  }
}

void t_scroll_reverse_one(buf)
     buffer *buf;
{
  /* scroll the screen down by one */
  if (set_scroll_region) {
    S = my_tsub(set_scroll_region, buf->startline,
		buf->height+buf->startline-1);
    TPUTS(S);
    free(S);
    move_cursor(0, buf->height+buf->startline-1);
    TPUTS(scroll_reverse);
    S = my_tsub(set_scroll_region, 0, rows-1);
    TPUTS(S);
    free(S);
    update_cursor(buf);
    buf->row++;
  } else {
/***/
    
  }
}
