/*
 * umud.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "shockit.h"
#include <stdio.h>
#include <time.h>

#define NUM_SOCKETS 10
#define UMUD_NAME_LEN 10

/* we'll wait 15 minutes before dying */
#define TIMEOUT (60*15)

void
  umud_tell(),
  umud_connect(),
  umud_disconnect(),
  umud_get_line(),
  umud_rename(),
  umud_WHO(),
  umud_time();

struct umud_conn {
  Socket         socket;
  char           name[UMUD_NAME_LEN+1];
  time_t         connect_time;
  time_t         last_act;
};

enum umud_ct { umud_inet, umud_unix };

enum umud_ct     umud_conn_type;
Socket           server;
struct umud_conn conn[NUM_SOCKETS];
int              in_use; /* number of sockets in use */
int              num_connections; /* number of connections
				   * that have been made */
int              umud_shutdown;
time_t           start_time, t;

void main(argc, argv)
     int argc;
     char *argv[];
{
  int fds[NUM_SOCKETS+1];
  char ready[NUM_SOCKETS+1];

  int x;

  if(argc < 2) {
    printf("usage: %s [port #|pathname]\n", argv[0]);
    exit(1);
  }

  if (isdigit(argv[1][0])) {
    umud_conn_type = umud_inet;
    sock_make_server(&server, NULL, atoi(argv[1]));
  } else {
    umud_conn_type = umud_unix;
    sock_make_server(&server, argv[1], UNIXDOM);
  }

  start_time = t = time(NULL);
  umud_shutdown = 0;
  in_use = 0;

  while(!umud_shutdown && (!TIMEOUT || (time(NULL)-t < TIMEOUT))) {
    /*
     * get rid of all unconnected sockets
     */
    for (x=0; x<in_use; x++)
      if (!sock_connected(conn[x].socket))
	umud_disconnect(x);

    /*
     * check all sockets to see which ones are waiting to spew data
     */
    for (x=0; x<in_use; x++)
      fds[x] = sock_fd(conn[x].socket);
    fds[in_use] = sock_fd(server);

    if (sock_select(fds, in_use+1, ready, 60000) > 0) {
      /*
       * check for new connections first
       */
      if (ready[in_use])
	umud_connect();

      /*
       * check for incoming lines
       */
      for (x=0; x<in_use; x++)
	if (ready[x])
	  umud_get_line(x);
    }
  }

  umud_tell(-1, "<%s> shutting down.\r\n", "");
}

/* send a line to all open connections */
void umud_tell(player, line, arg1)
     int player;
     char *line, *arg1;
{
  int x;

  for (x=0; x<in_use; x++)
    sock_printf(conn[x].socket, line,
		(player>=0) ? conn[player].name : "server", arg1);
}

void umud_connect()
{
  char buf[999];
  FILE *welcome;

  if (in_use == NUM_SOCKETS)
    return;

  if (!sock_accept(server, &(conn[in_use].socket), 100)
      && conn[in_use].socket) {
    conn[in_use].connect_time = conn[in_use].last_act = time(NULL);
    sprintf(conn[in_use].name, "%d", ++num_connections);
    in_use++;

    if ((welcome = fopen("hello.txt", "r")) != NULL) {
      while (fgets(buf, 999, welcome))
	sock_printf(conn[in_use-1].socket, "%s", buf);
      fclose(welcome);
    }

    umud_tell(in_use-1, "%s has connected\r\n", "");

    t = time(NULL);
  }
}

void umud_get_line(player)
     int player;
{
  static char *buff;
  static int bufflen=0;
  int len, readret;

  if (!bufflen)
    buff = (char *)malloc(bufflen=1000);

  strcpy(buff, "<%s> ");
  len = readret = sock_gets(conn[player].socket, buff+5, bufflen-5, 100);

  if (readret == 0)
    return;

  while ((readret != -1) && (buff[len+4] != '\n')) {
    buff = (char *)realloc(buff, bufflen+=500);
    readret = sock_gets(conn[player].socket, buff+5+len, bufflen-len-5, 100);
    len += readret;
  }

  t = conn[player].last_act = time(NULL);

  if (readret != -1) {
    if (!strcmp(buff+5, "QUIT\n"))
      umud_disconnect(player);
    else if (!strncmp(buff+5, "NAME ", 5))
      umud_rename(player, buff+10);
    else if (!strcmp(buff+5, "WHO\n"))
      umud_WHO(player);
    else if (!strcmp(buff+5, "!#%&(@$^*\n"))
      umud_shutdown = 1;
    else if (strcmp(buff+5, "\n"))
      umud_tell(player, buff, "");
  }
}

void umud_disconnect(player)
     int player;
{
  umud_tell(player, "%s has disconnected.\r\n", "");
  sock_close(&(conn[player].socket));
  free(conn[player].name);
  conn[player] = conn[--in_use];
}

void umud_rename(player, name)
     int player;
     char *name;
{
  int x;

  /* get rid of the trailing \n */
  name[strlen(name)-1] = '\0';
  if (strlen(name) == 0) {
    sock_printf(conn[player].socket,
		"You can't give yourself an empty name!\r\n");
    return;
  } else if (strlen(name) > UMUD_NAME_LEN) {
    sock_printf(conn[player].socket,
		"Name limited to %d characters. Truncating.\r\n", UMUD_NAME_LEN);
    name[UMUD_NAME_LEN] = '\0';
  }

  /* see if the name is already taken */
  for (x=0; x<in_use; x++)
    if ((x != player) && (!strcmp(conn[x].name, name))) {
      sock_printf(conn[player].socket,
		  "That name is already taken. Try another.\r\n");
      return;
    }

  umud_tell(player, "%s changes his name to %s\r\n", name);
  strcpy(conn[player].name, name);
}

void umud_WHO(player)
     int player;
{
  int x;
  char up[20], con[20], idl[20], name[1024];

  umud_time(start_time, up);
  sock_printf(conn[player].socket, "MicroMUD - up %s\r\n", up);

  switch (umud_conn_type) {
  case umud_inet:
    sock_printf(conn[player].socket, "Player Name    Connected Idle Host\r\n");
    sock_printf(conn[player].socket, "-----------    --------- ---- ----\r\n");
    for (x=0; x<in_use; x++) {
      umud_time(conn[x].connect_time, con);
      umud_time(conn[x].last_act, idl);
      if (sock_host(conn[x].socket, name) == -1)
	sock_ipaddr(conn[x].socket, name);
      sock_printf(conn[player].socket, "%-14s %9s %4s %s\r\n", conn[x].name,
		  con, idl, name);
    }
    break;
  case umud_unix:
    sock_printf(conn[player].socket, "Player Name    Connected Idle\r\n");
    sock_printf(conn[player].socket, "-----------    --------- ----\r\n");
    for (x=0; x<in_use; x++) {
      umud_time(conn[x].connect_time, con);
      umud_time(conn[x].last_act, idl);
      sock_printf(conn[player].socket, "%-14s %9s %4s\r\n", conn[x].name,
		  con, idl);
    }
    break;
  }
}

void umud_time(t, buf)
     time_t t;
     char *buf;
{
  int diff;

  diff = time(NULL) - t;

  if (diff < 60)
    sprintf(buf, "%ds", diff);
  else if (diff < 60*60)
    sprintf(buf, "%dm", diff/60);
  else if (diff < 60*60*24)
    sprintf(buf, "%dh", diff/(60*60));
  else
    sprintf(buf, "%dd", diff/(60*60*24));
}
