/*
 * stalk/var.c
 * Copyright (c) 1993 Gerald Weiler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "include.h"
#include "var.h"
#include "stuff.h"
#include "functions.h"
#include "debug.h"

#define CREATE_ERROR(x) \
  error_vars[(int) x] = newvar(); \
  error_vars[(int) x]->type = V_ERR; \
  error_vars[(int) x]->err = x;

static var *error_vars[NUM_ERRORS];
static char *error_codes[NUM_ERRORS];

/*
 * this is to keep a pool of allocated var structures for future need,
 * hopefully cutting down on a lot of allocating/deallocating.
 * (that seems to be 60% of the program time - at least before I did this
 * and put in var references)
 * if it runs out or overflows, the program will go back to malloc
 * and free
 */
#define POOL_SIZE 1000
static var *pool[POOL_SIZE];
static int pool_entries=0;

var_env *set_var(venv, name, value)
     var_env *venv;
     char *name;
     var *value;
{
  int c;
  var_env *t, *t2;

  if (venv == NULL) {
    t2 = t = NEW(var_env);
    t->name = mystrdup(name);
    t->value = var_ref(value);
    t->left = t->right = NULL;
    return t;
  }
  if (!(c = strcmp(venv->name, name))) {
    var_free(venv->value);
    venv->value = var_ref(value);
  } else if (c<0)
    venv->right = set_var(venv->right, name, value);
  else
    venv->left = set_var(venv->left, name, value);
  return venv;
}

var *get_var(venv, name)
     var_env *venv;
     char *name;
{
  int c;

  if (venv==NULL)
    return make_error(NULL, E_VARNF);

  if (!(c = strcmp(venv->name, name)))
    return var_ref(venv->value);
  else if (c<0)
    return get_var(venv->right, name);
  else
    return get_var(venv->left, name);
}

var *var_ref(value)
     var *value;
{
  int x;

  if (value != NULL) {
    value->ref++;
    if (value->type == V_LIST)
      for (x=0; x < value->num; x++)
	(void) var_ref(value->list[x]);
  }

  return value;
}

var *var_dup(value)
     var *value;
{
  int x;
  var *temp;

  if (value == NULL)
    return NULL;

  switch (value->type) {
  case V_STRING:
    return newstring(mystrdup(value->string));
    break;
  case V_LABEL:
    return newlabel(mystrdup(value->string));
    break;
  case V_LIST:
    temp = newlist(value->num);
    for (x=0; x<temp->num; x++)
      temp->list[x] = var_dup(value->list[x]);
    return temp;
    break;
  case V_NUM:
    return newnum(value->num);
    break;
  case V_ERR:
    return make_error(NULL, value->err);
    break;
  default:
    return NULL;
    break;
  }
}

void var_free(value)
     var *value;
{
  int x;

  if (value == NULL)
    return;

  if (value->type == V_LIST)
    for (x=0; x<value->num; x++)
      var_free(value->list[x]);

  value->ref--;
  if (value->ref > 0)
    return;
  else if (value->ref < 0)
    die("var.c, line %d: var reference == %d\n", __LINE__, value->ref);

  if ((value->type == V_LIST) && value->num)
    free(value->list);

  if ((value->type == V_STRING) || (value->type == V_LABEL))
    if (value->string)
      free(value->string);

  /*
   * put it in the pool if there's space. free it if not.
   * this will cut down on a lot of the malloc/free calls
   */
  if (pool_entries < POOL_SIZE)
    pool[pool_entries++] = value;
  else
    free(value);
}

void free_env(venv)
     var_env *venv;
{
  if (venv == NULL)
    return;
  free_env(venv->left);
  free_env(venv->right);
  var_free(venv->value);
  free(venv->name);
  free(venv);
}

void free_program(p)
     program *p;
{
  if (p == NULL)
    return;

  free(p->name);
  free_statements(p->base);
  free(p);
}

void free_statements(s)
     statement *s;
{
  statement *s1, *s2;

  s1 = s;

  while (s1) {
    s2 = s1;
    s1 = s1->next;
    var_free(s2->value);
    free(s2);
  }
}

char *err2string(err)
     Error err;
{
  return error_codes[(int) err];
}

int istrue(v)
     var *v;
{
  switch (v->type) {
  case V_STRING:
    return (strlen(v->string) != 0);
  case V_NUM:
    return (v->num != 0);
  case V_LIST:
    return (v->num > 0);
  case V_ERR:
    return 0;
  default:
    return 0;
  }
}

/*
 * if t is not NULL, make_error will raise an exception on task t.
 * since this could happen in the middle of some operation (esp a builtin
 * function), we will simply mark the task TA_DEBUG and put the error code
 * in the task structure. when the operation finishes, the task will be
 * debugged.
 */
var *make_error(t, err)
     task *t;
     Error err;
{
  if (t)
    raise_error(t, err);

  return var_ref(error_vars[(int) err]);
}

var *newvar()
{
  var *v;

  /* get a var from the pool if one's available. malloc if not */
  if (pool_entries)
    v = pool[--pool_entries];
  else
    v = NEW(var);
  v->type = V_NONE;
  v->ref = 1;
  v->string = NULL;
  v->list = NULL;
  return v;
}

var *newnum(n)
     int n;
{
  var *v;

  v = newvar();
  v->type = V_NUM;
  v->num = n;
  return v;
}

var *newtoken(m)
     mnemonic m;
{
  var *v;

  v = newvar();
  v->type = V_TOKEN;
  v->num = m;
  return v;
}

var *newstring(s)
     char *s;
{
  var *v;

  v = newvar();
  v->type = V_STRING;
  v->string = s;
  return v;
}

var *newlist(n)
     int n;
{
  var *v;
  int x;

  v = newvar();
  v->type = V_LIST;
  v->num = n;
  if (n)
    v->list = (var **)malloc(v->num * sizeof(var *));
  for (x=0; x<n; x++)
    v->list[x] = NULL;
  return v;
}

var *newlabel(s)
     char *s;
{
  var *v;

  v = newvar();
  v->type = V_LABEL;
  v->string = s;
  return v;
}

int equal_vars(v1, v2)
     var *v1, *v2;
{
  int x;

  if (v1->type != v2->type)
    return 0;
  if (v1->type == V_NUM)
    return (v1->num == v2->num);
  else if (v1->type == V_ERR)
    return (v1->err == v2->err);
  else if (v1->type == V_STRING)
    return !mystrcasecmp(v1->string, v2->string);
  else if (v1->type != V_LIST)
    return 0;

  /* check to see if the lists are the same recursively */
  if (v1->num != v2->num)
    return 0;

  for (x=0; x<v1->num; x++)
    if (!equal_vars(v1->list[x], v2->list[x]))
      return 0;

  return 1;
}

char *vartostr(v)
     var *v;
{
  char *s, *t1, *t2;
  int x;

  if (v == NULL)
    return mystrdup("");
  switch(v->type) {
  case V_STRING:
  case V_LABEL:
    return mystrdup(v->string);
    break;
  case V_ERR:
    return mystrdup(err2string(v->err));
    break;
  case V_NUM:
  case V_INDEX:
    s = (char *)malloc(20);
    sprintf(s, "%d", v->num);
    return s;
    break;
  case V_LIST:
    if (v->num == 0)
      return mystrdup("");
    t1 = vartostr(v->list[0]);
    for (x=1; x<v->num; x++) {
      t2 = vartostr(v->list[x]);
      s = malloc(strlen(t1) + strlen(t2) + 2);
      sprintf(s, "%s%s", t1, t2);
      free(t1);
      free(t2);
      t1 = s;
    }
    return t1;
    break;
  case V_TOKEN:
    die("var.c: invalid var type TOKEN, line %d\n", __LINE__);
    break;
  default:
    die("var.c: invalid var type %d, line %d\n", v->type, __LINE__);
    break;
  }
  return NULL;
}

Error inv_args(args, min, max, types)
     var *args;
     int min, max;
     var_type types[];
{
  int x;

  if ((min == 0) && !args)
    return E_NONE;

  if (!args || (args->type != V_LIST))
    return E_TYPE;

  if ((args->num < min) || (args->num > max))
      return E_ARGS;

  for (x=0; x<args->num; x++)
    if (args->list[x]->type != types[x])
      return E_TYPE;

  return E_NONE;
}

var *varlist(t)
     task *t;
{
  var *r;

  r = newlist(2);
  r->list[0] = varlist_sub(t->local);
  r->list[1] = varlist_sub(t->buf->global);
  return r;
}

var *varlist_sub(v)
     var_env *v;
{
  var *l, *r, *t;
  int i, j;

  if (!v)
    return NULL;

  l = varlist_sub(v->left);
  r = varlist_sub(v->right);
  t = newlist((l ? l->num : 0) + (r ? r->num : 0) +1);
  j = 0;
  if (l) {
    for (i=0; i<l->num; i++)
      t->list[i] = var_ref(l->list[i]);
    var_free(l);
    j = l->num;
  }
  t->list[j++] = newstring(v->name);
  if (r) {
    for (i=0; i<r->num; i++)
      t->list[i+j] = var_ref(r->list[i]);
    var_free(r);
  }
  return t;
}

var *extract_elements(t, v, i1, i2)
     task *t;
     var *v;
     int i1, i2;
{
  var *result;
  int x;

  /*
   * i1 and i2 are the indices.
   * if i2 == -1, then we want a single element: v[i1]
   * if not, we want v1[i1..i2]
   */
  if (v->type == V_STRING) {
    if (i2 == -1)
      i2 = i1;
    if (i1 > i2)
      return make_error(t, E_INVARG);
    if  ((i1 < 1) || (i2 > strlen(v->string)))
      return make_error(t, E_RANGE);

    result = newstring((char *)malloc(i2-i1+2));
    strncpy(result->string, v->string+i1-1, i2-i1+1);
    result->string[i2-i1+1] = '\0';
  } else if (v->type == V_LIST) {
    if (i2 == -1) {
      if ((i1 < 1) || (i1 > v->num))
	return make_error(t, E_RANGE);
      return var_ref(v->list[i1-1]);
    }
    if (i1 > i2)
      return make_error(t, E_INVARG);
    if  ((i1 < 1) || (i2 > v->num))
      return make_error(t, E_RANGE);

    result = newlist(i2 - i1 + 1);
    for (x=0; x<result->num; x++)
      result->list[x] = var_ref(v->list[i1-1+x]);
  } else
    return make_error(t, E_TYPE);

  return result;
}

void init_vars()
{
  error_codes[E_NONE]   = "No error";
  error_codes[E_VARNF]  = "Variable not found";
  error_codes[E_TYPE]   = "Type mismatch";
  error_codes[E_DIV]    = "Division by zero";
  error_codes[E_VERBNF] = "Verb not found";
  error_codes[E_MAXREC] = "Too many function calls";
  error_codes[E_RANGE]  = "Range error";
  error_codes[E_ARGS]   = "Incorrect number of arguments";
  error_codes[E_INVARG] = "Invalid argument";
  error_codes[E_EXIT]   = "Exit not in loop structure";
  error_codes[E_TICKS]  = "Task ran out of ticks";

  CREATE_ERROR(E_NONE);
  CREATE_ERROR(E_VARNF);
  CREATE_ERROR(E_TYPE);
  CREATE_ERROR(E_DIV);
  CREATE_ERROR(E_VERBNF);
  CREATE_ERROR(E_MAXREC);
  CREATE_ERROR(E_RANGE);
  CREATE_ERROR(E_ARGS);
  CREATE_ERROR(E_INVARG);
  CREATE_ERROR(E_EXIT);
  CREATE_ERROR(E_TICKS);
}

char *err2token(err_val)
     Error err_val;
{
  switch (err_val) {
  case E_NONE:
    return "E_NONE";
    break;
  case E_VARNF:
    return "E_VARNF";
    break;
  case E_TYPE:
    return "E_TYPE";
    break;
  case E_DIV:
    return "E_DIV";
    break;
  case E_VERBNF:
    return "E_VERBNF";
    break;
  case E_MAXREC:
    return "E_MAXREC";
    break;
  case E_RANGE:
    return "E_RANGE";
    break;
  case E_ARGS:
    return "E_ARGS";
    break;
  case E_INVARG:
    return "E_INVARG";
    break;
  case E_EXIT:
    return "E_EXIT";
    break;
  case E_TICKS:
    return "E_TICKS";
    break;
  default:
    return "";
    break;
  }
}
